import torch
import torch.nn as nn
import math


class MogrifierLSTMCell(nn.Module):

    def __init__(self, input_size, hidden_size, mogrify_steps):
        super(MogrifierLSTMCell, self).__init__()
        self.hidden_size = hidden_size
        self.input_size = input_size
        self.x2h = nn.Linear(input_size, 4 * hidden_size)
        self.h2h = nn.Linear(hidden_size, 4 * hidden_size)
        self.q = nn.Linear(hidden_size, input_size)
        self.r = nn.Linear(input_size, hidden_size)
        self.tanh = nn.Tanh()
        self.init_parameters()
        self.mogrify_steps = mogrify_steps

    def init_parameters(self):
        std = 1.0 / math.sqrt(self.hidden_size)
        for p in self.parameters():
            p.data.uniform_(-std, std)

    def mogrify(self, x, h):
        for i in range(1, self.mogrify_steps + 1):
            if i % 2 == 0:
                h = (2 * torch.sigmoid(self.r(x))) * h
            else:
                x = (2 * torch.sigmoid(self.q(h))) * x
        return x, h

    def forward(self, x, states):
        """
        inp shape: (batch_size, input_size)
        each of states shape: (batch_size, hidden_size)
        """
        ht, ct = states
        x, ht = self.mogrify(x, ht)  # Note: This should be called every timestep
        gates = self.x2h(x) + self.h2h(ht)  # (batch_size, 4 * hidden_size)
        in_gate, forget_gate, new_memory, out_gate = gates.chunk(4, 1)
        in_gate = torch.sigmoid(in_gate)
        forget_gate = torch.sigmoid(forget_gate)
        out_gate = torch.sigmoid(out_gate)
        new_memory = self.tanh(new_memory)
        c_new = (forget_gate * ct) + (in_gate * new_memory)
        h_new = out_gate * self.tanh(c_new)

        return h_new, c_new

class MogrifierLSTM(torch.nn.Module):
    def __init__(self, input_size,  hidden_size, mogrify_steps=2, learn_initial_states=False, batch_first=True):
        super(MogrifierLSTM, self).__init__()
        self.cell = MogrifierLSTMCell(input_size, hidden_size, mogrify_steps)

        self.hidden_size = hidden_size
        self.input_size = input_size
        self.batch_first=batch_first

        self.h = torch.nn.Parameter(torch.zeros(1, self.hidden_size))
        self.c = torch.nn.Parameter(torch.zeros(1, self.hidden_size))
        self.h.requires_grad = learn_initial_states
        self.c.requires_grad = learn_initial_states

    def forward(self, x):
        if not self.batch_first:
            x = x.permute(1,0,2)

        max_len=x.shape[1]
        batch_size=x.shape[0]
        hidden_states = []
        h = self.h.repeat(batch_size,1)
        c = self.c.repeat(batch_size,1)

        for step in range(max_len):
            t = x[:, step]
            h, c = self.cell(t, (h, c))
            hidden_states.append(h.unsqueeze(1))

        return torch.cat(hidden_states, dim = 1), (h, c)


class MogLSTM(nn.Module):
    def __init__(self, input_sz: int, hidden_sz: int, mog_iterations: int):
        super().__init__()
        self.input_size = input_sz
        self.hidden_size = hidden_sz
        self.mog_iterations = mog_iterations
        # Define/initialize all tensors
        self.Wih = torch.nn.Parameter(torch.Tensor(input_sz, hidden_sz * 4))
        self.Whh = torch.nn.Parameter(torch.Tensor(hidden_sz, hidden_sz * 4))
        self.bih = torch.nn.Parameter(torch.Tensor(hidden_sz * 4))
        self.bhh = torch.nn.Parameter(torch.Tensor(hidden_sz * 4))
        # Mogrifiers
        self.Q = torch.nn.Parameter(torch.Tensor(hidden_sz, input_sz))
        self.R = torch.nn.Parameter(torch.Tensor(input_sz, hidden_sz))

        self.init_weights()

    def init_weights(self):
        for p in self.parameters():
            if p.data.ndimension() >= 2:
                nn.init.xavier_uniform_(p.data)
            else:
                nn.init.zeros_(p.data)
        self.Whh = nn.init.orthogonal_(self.Whh)

        self.bih.data[self.hidden_size:2 * self.hidden_size] = 1
        self.bhh.data[self.hidden_size:2 * self.hidden_size] = 1

    def mogrify(self, xt, ht):
        for i in range(1, self.mog_iterations + 1):
            if (i % 2 == 0):
                ht = (2 * torch.sigmoid(xt @ self.R)) * ht
            else:
                xt = (2 * torch.sigmoid(ht @ self.Q)) * xt
        return xt, ht

    # Define forward pass through all LSTM cells across all timesteps.
    # By using PyTorch functions, we get backpropagation for free.
    def forward(self, x: torch.Tensor, init_states = None):
        """Assumes x is of shape (batch, sequence, feature)"""
        batch_sz, seq_sz, _ = x.size()
        hidden_seq = []
        # ht and Ct start as the previous states and end as the output states in each loop below
        if init_states is None:
            ht = torch.zeros((batch_sz, self.hidden_size)).to(x.device)
            Ct = torch.zeros((batch_sz, self.hidden_size)).to(x.device)
        else:
            ht, Ct = init_states
        for t in range(seq_sz):  # iterate over the time steps
            xt = x[:, t, :]
            xt, ht = self.mogrify(xt, ht)  # mogrification
            gates = (xt @ self.Wih + self.bih) + (ht @ self.Whh + self.bhh)
            ingate, forgetgate, cellgate, outgate = gates.chunk(4, 1)

            ### The LSTM Cell!
            ft = torch.sigmoid(forgetgate)
            it = torch.sigmoid(ingate)
            Ct_candidate = torch.tanh(cellgate)
            ot = torch.sigmoid(outgate)
            # outputs
            Ct = (ft * Ct) + (it * Ct_candidate)
            ht = ot * torch.tanh(Ct)
            ###

            hidden_seq.append(ht.unsqueeze(0))
        hidden_seq = torch.cat(hidden_seq, dim=0)
        # reshape from shape (sequence, batch, feature) to (batch, sequence, feature)
        hidden_seq = hidden_seq.transpose(0, 1).contiguous()
        return hidden_seq, (ht, Ct)



"""
Defines dataset class and provides automated load/cache functions for some public multilabel datasets.
"""

from torch.utils.data import Dataset
import torch
from .data_loaders  import load_eurlex, load_wiki30k, load_huffpost, load_aapd, load_rcv1, \
    load_moviesummaries,load_blurbgenrecollection, load_blurbgenrecollection_de, load_20newsgroup,export,\
    load_agnews, load_dbpedia, load_ohsumed

# String Mappings
register = {
    "aapd": load_aapd,
    "rcv1": load_rcv1,
    "huffpost": load_huffpost,
    "wiki30k": load_wiki30k,
    "eurlex": load_eurlex,
    "movies_summaries": load_moviesummaries,
    "blurbgenrecollection": load_blurbgenrecollection,
    "blurbgenrecollection_de": load_blurbgenrecollection_de,
    "20newsgroup": load_20newsgroup,
    "agnews": load_agnews,
    "dbpedia": load_dbpedia,
    "ohsumed": load_ohsumed
}



class MultiLabelDataset(Dataset):
    """
    Dataset to hold text and label combinations.

    Providing a unified interface to Multilabel data, associating textual input with sets of labels. Also
    holding a mapping of labels to indices and transforming the target labelset of an instance to multi-hot representations.
    It also inherits torch.utils.data.Dataset, so it can be used in combination with torch.utils.data.Dataloader
    for fast training loops.
    """
    def __init__(self, x, y, classes, target_dtype=torch._cast_Float, one_hot=True, **kwargs):
        """
        Class constructor

        Creates an instance of MultilabelDataset.

        Args:
            x: A list of the input text
            y:  A list of corresponding label sets
            classes: A class mapping from label strings to successive indices
            target_dtype: The final cast on the label output. (Some of torch's loss functions expect other data types. This argument defines
                a function that is applied to the final output of the label tensors. (default: torch._cast_Float)
            one_hot: (default: True) if True, will transform the multilabel sets into a multi-hot tensor representation for training
                if False: will return the labelset strings as is
            **kwargs:  Any additional information that is given by named keywords will be saved as metadata

        Returns:
            A MultilabelDataset instance.

        Examples:
            ```
            x = ["This is a text about science and philosophy",
                "This is another text about science and politics"]


            y = [['science', 'philosophy'],
                ['science', 'politics']]

            classes = {
                "science": 0,
                "philosophy": 1,
                "politics": 2
            }
            dataset = mlmc.data.MultilabelDataset(x=x, y=y, classes=classes)
            dataset[0]
            ```

        """

        self.__dict__.update(kwargs)
        self.classes = classes
        self.x = x
        self.y = y
        self.one_hot = one_hot
        self.target_dtype = target_dtype

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        if self.one_hot:
            labels = [self.classes[tag] for tag in self.y[idx]]
            labels = torch.nn.functional.one_hot(torch.LongTensor(labels), len(self.classes)).sum(0)
            return {'text': self.x[idx], 'labels': self.target_dtype(labels)}
        else:
            return {'text': self.x[idx], 'labels': self.y[idx]}

    def transform(self, fct):
        """
        Mapping functions that act on strings to every data instance

        Applies fct to every input element of the dataset. (Can be used for cleaning or preprocessing)

        Args:
            fct: A function that takes a string as input and returns the transformed string
        """
        self.x = [fct(sen) for sen in self.x]

    def to_json(self):
        """
        Transform the data set into a json string representation

        Returns:
             String representation of the dataset ( only x, y and the classes)
        """
        import json
        json_string = json.dumps(self.to_dict())
        return json_string

    def to_dict(self):
        """
        Transform the dataset into a dictionary-of-lists representation

         Returns:
              A python dictionary of the training data (only x, y and the classes)
        """
        return {"x": self.x, "y": self.y, "classes":list(self.classes.keys())}

    def __add__(self, o):
        new_classes = list(set( list(self.classes.keys()) +  list(o.classes.keys()) ))
        new_classes.sort()
        new_classes = dict(zip(new_classes, range(len(new_classes))))

        new_data = list(set(self.x + o.x))
        new_labels = [[] for _ in range(len(new_data))]

        for i, x in enumerate(self.x):
            new_labels[new_data.index(x)].extend(self.y[i])

        for i, x in enumerate(o.x):
            new_labels[new_data.index(x)].extend(o.y[i])

        new_labels = [list(set(x)) for x in new_labels]

        return MultiLabelDataset(x=new_data, y=new_labels, classes=new_classes)

    def remove(self, classes):
        """
        Deleting labels from the dataset.

        Removes all occurrences of classes argument (string or list of strings) from the dataset.
        Instances with then empty labelsets will be removed completely

        Args:
            classes: A label or list of label names.
        """
        if isinstance(classes, str):
            classes = [classes]
        assert all([x in self.classes.keys() for x in classes]), "Some of the provided classes are not contained in the dataset"
        self.y = [[label for label in labelset if label not in classes] for labelset in self.y]
        emptylabelsets = [i for i, x in enumerate(self.y) if x == []]
        self.x = [ x for i,x in enumerate(self.x) if i not in emptylabelsets]
        self.y = [ x for i,x in enumerate(self.y) if i not in emptylabelsets]

    def map(self, map: dict):
        """
        Transforming label names

        Apply label mappings to every data instance. Maps every label string in the dataset according to 'map'.

        Args:
            map: Dictionary of map from current label string to new label string
        """
        if any([x not in map.keys() for x in self.classes.keys()]):
            print("Some classes are not present in the map. The will be returned as is.")
        self.classes = {map.get(k,k):v for k,v in self.classes.items()}
        self.y = [[map.get(l,l) for l in labelset] for labelset in self.y]

    def reduce(self, subset: dict):
        """
        Reduces the dataset to a subset of the classes.

        The resulting dataset will only contain instances with at least one label that appears in the subset argument.
        The subset must also provide a new mapping from the new label names to indices.
        All labels not in subset will be removed. Instances with an empty label set will be removed.

        Args:
            subset: A mapping of classes to indices
        """
        assert all([x in self.classes.keys() for x in subset.keys()]), "Subset contains classes not present in dataset"
        ind = [i for i, labelset in enumerate(self.y) if any([l in subset.keys() for l in labelset])]
        self.x = [self.x[i] for i in ind]
        self.y = [[x for x in self.y[i] if x in subset.keys()] for i in ind]
        self.classes = subset

    def count(self, label):
        """
        Count the occurrences of all labels in 'label' in the dataset.

        Args:
            label: Label name or list of label names
        Returns:
             Dictionary of label name and frequency in the dataset.
        """
        if isinstance(label, list):
            result = {   l: sum([l in s for s in self.y])
                for l in label
            }
        else:
            result = {label: sum([label in s for s in self.y])}
        return result

    def density(self):
        """
        Returns the average label set size per instance.

        Returns: The average labelset size per instance
        """
        return sum([len(x) for x in self.y])/len(self.y)


class SingleLabelDataset(MultiLabelDataset):
    def __init__(self, *args, **kwargs):
        super(SingleLabelDataset, self).__init__(*args, **kwargs)
        assert all([len(x)==1 for x in self.y]), "This is not a single label dataset. Some labels contain multiple labels."

    def __getitem__(self, idx):
        return {'text': self.x[idx], 'labels': torch.tensor(self.classes[self.y[idx][0]])}



#-------------------------------------------------------------------------------------


def get_dataset(name, type, ensure_valid=False, valid_split=0.25, target_dtype=torch.FloatTensor):
    """
    General dataset getter for datasets in provided by the package.

    :param name: name of the dataset in register
    :param type: MultilabelDataset or SequenceDataset defined in mlmc.data
    :param ensure_valid: if True and there's no validation data in the original data a portion of the trainset is split
    :param valid_split: the fraction of the train set to be used as validation if ensure_valid=True
    :param target_dtype: Target Tensortype of the label multihot representation. (default torch.FloatTensor)
    :return: a dictionary with keys: "train", "valid" and "test" and additional information the dataset provides (graphs, maps, classes,..)
    """
    data, classes = register.get(name, None)()
    if data is None:
        Warning("data not found")
        return None
    else:
        if "valid" not in data.keys():
            print("No Validation data found.")
            if ensure_valid:
                print("Providing random split...")
                from sklearn.model_selection import train_test_split
                splits_from_train = train_test_split(*data["train"], test_size=valid_split)
                train=[splits_from_train[0],splits_from_train[2]]
                valid = [splits_from_train[1],splits_from_train[3]]
                data["train"]=train
                data["valid"]=valid
            else:
                data["valid"]= None
        datasets = {
            split: type(x=data[split][0],
                                     y=data[split][1],
                                     classes=classes,
                                     purpose=split,
                                     target_dtype=target_dtype) if data[split] is not None else None
        for split in ["test","train","valid"]
        }
        for k in data.keys():
            if k not in ["test", "train", "valid"]:
                datasets[k] = data[k]
        datasets["classes"]=classes

        return datasets

## Wrapper for multilabel datasets
def get_multilabel_dataset(name, target_dtype=torch._cast_Float):
    """
    Load multilabel training data if available.

    This is the default wrapper function for retrieving multilabel datasets.

    :param name: See: mlmc.data.register.keys()
    :param target_dtype: The target_dtype of the labeldata in training. See MultilabelDataset
    :return:
    """
    return get_dataset(name, type=MultiLabelDataset, ensure_valid=False, target_dtype=torch._cast_Float)


## Wrapper for singlelabel datasets
def get_singlelabel_dataset(name, target_dtype=torch._cast_Float):
    """
    Load singlelabel training data if available.

    This is the default wrapper function for retrieving SingleLabelDataset datasets.
    This is a special case of MultilabelDataset.

    :param name: See: mlmc.data.register.keys()
    :param target_dtype: The target_dtype of the labeldata in training. See MultilabelDataset
    :return:
    """
    return get_dataset(name, type=SingleLabelDataset, ensure_valid=False, target_dtype=torch._cast_Float)




## Sampler import
from .sampler import sampler, successive_sampler, class_sampler, validation_split
from .data_loaders_text import RawTextDatasetTokenizer, RawTextDataset,RawTextDatasetTensor



from sklearn.decomposition import PCA
import numpy as np

def postprocess_embedding(x, D=10):
    """Remove principal components and mean value to make embeddings more discernable
    Mu and Visanath (2018): https://arxiv.org/pdf/1702.01417.pdf"""
    no_mean = x-x.mean(0)[None,:]
    pca = PCA(n_components=D, svd_solver='full')
    pca.fit(no_mean)
    return no_mean - np.matmul(np.matmul(pca.components_,x.transpose()).transpose(), pca.components_)




import torch
from ..data.transformer import label_smoothing_random


class NoiseSmoothBCEWithLogitsLoss(torch.nn.BCEWithLogitsLoss):
    """Torch like loss function but with smoothed labels. [Confident Learning]
    ToDo:
     - Make the smoothing function an argument
    """
    def __init__(self):
        super(NoiseSmoothBCEWithLogitsLoss, self).__init__()
        self.smoothing_noise = 0.2

    def forward(self, inputs, targets):
        loss = super(NoiseSmoothBCEWithLogitsLoss,self).forward(label_smoothing_random(inputs, self.smoothing_noise), targets)
        return loss



import numpy as np
import itertools
from sklearn.manifold import TSNE


def cooc_matrix(labels, classes):
    """Deprecated"""
    coocs = np.zeros((len(classes),len(classes)))
    frequencies = np.zeros((len(classes),1))
    for labelset in labels:
        for p in list(itertools.combinations(labelset, 2)):
            coocs[classes[p[0]],classes[p[1]]] += 1
            coocs[classes[p[1]],classes[p[0]]] += 1
            frequencies[classes[p[1]],0] +=1
            frequencies[classes[p[0]], 0] += 1

    return coocs, frequencies


def correlate_similarity(coocs, embeddings, n, classwise=False, corr="spearman"):
    """Deprecated"""
    cooc_rank = np.argsort(coocs, -1)[:, -n::-1]
    embed_rank = np.argsort(np.dot(embeddings, embeddings.transpose()), -1, )[:, -n::-1]

    from scipy.stats import spearmanr, kendalltau
    classcorrelations = []
    if corr=="spearman": fct = spearmanr
    if corr=="kendalltau": fct = kendalltau
    for  a, b in zip(cooc_rank, embed_rank):
        classcorrelations.append(fct(a, b)[0])

    if classwise:
        return classcorrelations
    else:
        return [np.mean(np.abs(classcorrelations)), np.std(np.abs(classcorrelations))],\
               [np.mean(classcorrelations), np.std(classcorrelations)]

def show_graph(vectors, classes):
    """
    Plot A TSNE
    :param vectors: Vectors to plot
    :param classes: Labels of the vectors to plot
    :return:
    """
    import matplotlib.pyplot as plt
    X_embedded = TSNE(n_components=2).fit_transform(vectors)
    fig, ax = plt.subplots(figsize=(20, 20))
    ax.scatter(X_embedded[:, 0], X_embedded[:, 1])
    for i, txt in enumerate(classes.keys()):
        ax.annotate(txt, (X_embedded[i, 0], X_embedded[i, 1]))
    plt.show()



import torch

class Metric(torch.nn.Module):
    """Experimental feature A simple weighted vector sum.
        Idea:
            x[i, j] every activation i of the input corresponds to a distribution over the output class j
            The distributions are weighted and summed.
    """
    def __init__(self, dim):
        super(Bilinear, self).__init__()
        self.dim = dim
        self.metric_tensor = torch.nn.Parameter(
            torch.triu(torch.zeros((dim, dim)))
        )
        self.metric_tensor.requires_grad=True
        torch.nn.init.eye_(self.metric_tensor)


    def forward(self, x, y):
        assert len(y.shape)==2, "y has to a"
        metric = torch.matmul(self.metric_tensor.triu(), self.metric_tensor.triu().t())
        # differences = (x.unsqueeze(-1) - y.t().unsqueeze(0)).permute(0,2,1)
        # p = torch.matmul(differences, metric)
        # p = (differences*p).sum(-1)
        return torch.matmul(torch.matmul(x,metric.t()),y.t())

    def regularize(self):
        lower_tri = torch.triu(self.metric_tensor).t()
        dev = torch.relu((lower_tri.sum(-1) - 2 * lower_tri.diag()))
        return dev.sum() + torch.relu(-lower_tri.diag()).sum()



class Bilinear(torch.nn.Module):
    """Experimental feature A simple weighted vector sum.
        Idea:
            x[i, j] every activation i of the input corresponds to a distribution over the output class j
            The distributions are weighted and summed.
    """
    def __init__(self, dim):
        super(Bilinear, self).__init__()
        self.dim = dim
        self.metric_tensor = torch.nn.Parameter(
            torch.triu(torch.zeros((dim, dim)))
        )
        self.metric_tensor.requires_grad=True
        torch.nn.init.eye_(self.metric_tensor)


    def forward(self, x, y):
        assert len(y.shape)==2, "y has to have 2 dimensions"
        metric = torch.matmul(self.metric_tensor.triu(), self.metric_tensor.triu().t())
        return torch.matmul(torch.matmul(x,metric.t()),y.t())

    def regularize(self):
        lower_tri = torch.triu(self.metric_tensor).t()
        dev = torch.relu((lower_tri.sum(-1) - 2 * lower_tri.diag()))
        return dev.sum() + torch.relu(-lower_tri.diag()).sum()


class Metric(torch.nn.Module):
    """Experimental feature A simple weighted vector sum.
        Idea:
            x[i, j] every activation i of the input corresponds to a distribution over the output class j
            The distributions are weighted and summed.
    """
    def __init__(self, dim):
        super(Metric, self).__init__()
        self.dim = dim
        self.metric_tensor = torch.nn.Parameter(
            torch.triu(torch.zeros((dim, dim)))
        )
        self.metric_tensor.requires_grad=True
        torch.nn.init.eye_(self.metric_tensor)


    def forward(self, x, y):
        assert len(y.shape)==2, "y has to a"
        metric = torch.matmul(self.metric_tensor.triu(), self.metric_tensor.triu().t())
        n_unsqueezes = len(x.shape)-2
        m=y.t()
        for i in range(n_unsqueezes):
            m = m.unsqueeze(0)
        differences = (x.unsqueeze(-1) - m).permute(0,2,1)
        p = torch.matmul(differences, metric)
        p = (differences*p).sum(-1)
        return p

    def regularize(self):
        lower_tri = torch.triu(self.metric_tensor).t()
        dev = torch.relu((lower_tri.sum(-1) - 2 * lower_tri.diag()))
        return dev.sum() + torch.relu(-lower_tri.diag()).sum()




"""
Main Module import all submodules.
"""
#
#
#
__author__ = "Janos Borst"
try:
    from mlmc._version import __version__
except:
    __version__ = "unknown"

import mlmc.data
import mlmc.models
import mlmc.graph
import mlmc.metrics
import mlmc.representation
# Save and load models for inference
from .save_and_load import save, load


import torch
from tqdm import tqdm

from ..metrics.multilabel import MultiLabelReport, AUC_ROC
from ..representation import is_transformer, get
from ..representation.labels import makemultilabels

from ..representation import threshold_mcut, threshold_hard,threshold_max

from ..data import SingleLabelDataset,MultiLabelDataset

class TextClassificationAbstract(torch.nn.Module):
    """
    Abstract class for Multilabel Models. Defines fit, evaluate, predict and threshold methods for virtually any
    multilabel training.
    This class is not meant to be used directly.
    Also provides a few default functions:
        _init_input_representations(): if self.representations exists, the default will load a embedding and corresponding tokenizer
        transform(): If self.tokenizer exists the default method wil use this to transform text into the models input

    """
    def __init__(self, target="multi", activation=None, loss=None, optimizer=torch.optim.Adam, optimizer_params={"lr": 5e-5}, device="cpu", **kwargs):
        """
        Abstract initializer of a Text Classification network.
        Args:
            target: single label oder multilabel mode. defined by keystrings: ("single", "multi"). Sets some basic options, like loss function, activation and
                    metrics to sensible defaults.
            activation: The activation function applied to the output. Only used for metrics and when you want to return scores in predict. (default: torch.softmax for "single", torch.sigmoid for "multi")
            loss: One of the torch.nn  losses (default: torch.nn.BCEWithLogitsLoss for "multi" and torch.nn.CrossEntropyLoss for "single")
            optimizer:  One of toch.optim (default: torch.optim.Adam)
            optimizer_params: A dictionary of optimizer parameters
            device: torch device, destination of training (cpu or cuda:0)
        """
        super(TextClassificationAbstract,self).__init__(**kwargs)

        assert target in ("multi", "single"), 'target must be one of "multi" or "single"'

        # Setting default values for learning mode
        self.target = target
        if target == "single":
            self.activation = torch.softmax
            self.loss = torch.nn.CrossEntropyLoss
        elif self.target == "multi":
            self.activation = torch.sigmoid
            self.loss = torch.nn.BCEWithLogitsLoss

        # If there were external arguments we will use them
        if activation is not None:
            self.activation = activation
        if loss is not None:
            self.loss = loss

        self.device = device
        self.optimizer = optimizer
        self.optimizer_params = optimizer_params
        self.PRECISION_DIGITS = 4

    def act(self, x):
        if "softmax" in self.activation.__name__ or "softmin" in self.activation.__name__:
            return self.activation(x,-1)
        else:
            return self.activation(x)


    def build(self):
        """
        Internal build method.
        """
        if isinstance(self.loss, type) and self.loss is not None:
            self.loss = self.loss().to(self.device)
        if isinstance(self.optimizer, type) and self.optimizer is not None:
            self.optimizer = self.optimizer(filter(lambda p: p.requires_grad, self.parameters()), **self.optimizer_params)
        self.to(self.device)

    def evaluate_classes(self, classes_subset=None, **kwargs):
        """wrapper for evaluation function if you just want to evaluate on subsets of the classes."""
        if classes_subset is None:
            return self.evaluate(**kwargs)
        else:
            mask = makemultilabels([list(classes_subset.values())], maxlen=len(self.classes))
            return self.evaluate(**kwargs, mask=mask)

    def evaluate(self, data, batch_size=50, return_roc=False, return_report=False, mask=None):
        """
        Evaluation, return accuracy and loss and some multilabel measure

        Returns p@1, p@3, p@5, AUC, loss, Accuracy@0.5, Accuracy@mcut, ROC Values, class-wise F1, Precision and Recall.
        Args:
            data: A MultilabelDataset with the data for evaluation
            batch_size: The batch size of the evaluation loop. (Larger is often faster, but it should be small enough to fit into GPU memory. In general it can be larger than batch_size in training.
            return_roc: If True, the return dictionary contains the ROC values.
            return_report: If True, the return dictionary will contain a class wise report of F1, Precision and Recall.
        Returns:
            A dictionary with the evaluation measurements.
        """
        self.eval()  # set mode to evaluation to disable dropout
        from ignite.metrics import Average
        from ..metrics import PrecisionK, AccuracyTreshold


        assert not (type(data)== SingleLabelDataset and self.target=="multi"), "You inserted a SingleLabelDataset but chose multi as target."
        assert not (type(data) == MultiLabelDataset and self.target=="single"), "You inserted a MultiLabelDataset but chose single as target."

        multilabel_metrics = {
            "p@1": PrecisionK(k=1, is_multilabel=True, average=True),
            "p@3": PrecisionK(k=3, is_multilabel=True, average=True),
            "p@5": PrecisionK(k=5, is_multilabel=True, average=True),
            "tr@0.5": AccuracyTreshold(trf=threshold_hard, args_dict={"tr": 0.5}, is_multilabel=True),
            "mcut": AccuracyTreshold(trf=threshold_mcut, is_multilabel=True),
            "auc_roc": AUC_ROC(len(self.classes), return_roc=return_roc),
        }
        if return_report:
            multilabel_metrics["report"] = MultiLabelReport(self.classes, trf=threshold_mcut) \
                if mask is None else MultiLabelReport(self.classes, trf=threshold_mcut, check_zeros=True)

        if len(self.classes) <= 5:
            del multilabel_metrics["p@5"]
        if len(self.classes) <= 3:
            del multilabel_metrics["p@3"]

        singlelabel_metrics = {
            "accuracy":  AccuracyTreshold(threshold_max, is_multilabel=False)
        }

        metrics = multilabel_metrics
        if self.target == "single":
            metrics = singlelabel_metrics

        average = Average()
        data_loader = torch.utils.data.DataLoader(data, batch_size=batch_size)
        with torch.no_grad():
            for i, b in enumerate(data_loader):
                y = b["labels"]
                x = self.transform(b["text"])
                output = self(x.to(self.device)).cpu()
                if hasattr(self, "regularize"):
                    l = self.loss(output, y) + self.regularize()
                else:
                    l = self.loss(output, y)
                output = self.act(output)

                # Subset evaluation if ...
                if mask is not None:
                    output = output * mask
                    y = y * mask

                average.update(l.item())
                for v in metrics.values():
                    v.update((output, y))
        self.train()

        results = {"valid_loss": round(average.compute().item(), 2*self.PRECISION_DIGITS)}
        results.update({k: round(v.compute(), self.PRECISION_DIGITS) if isinstance(v.compute(), float) else v.compute() for k, v in metrics.items()})

        return results

    def fit(self, train,
            valid=None, epochs=1, batch_size=16, valid_batch_size=50, classes_subset=None, patience=-1, tolerance=1e-2,
            return_roc=False, return_report=False):
        """
        Training function

        Args:
            train: MultilabelDataset used as training data
            valid: MultilabelDataset to keep track of generalization
            epochs: Number of epochs (times to iterate the train data)
            batch_size: Number of instances in one batch.
            valid_batch_size: Number of instances in one batch  of validation.
            patience: (default -1) Early Stopping Arguments. Number of epochs to wait for performance improvements before exiting the training loop.
            tolerance: (default 1e-2) Early Stopping Arguments. Minimum improvement of an epoch over the best validation loss so far.

        Returns:
            A history dictionary with the loss and the validation evaluation measurements.

        """
        import datetime
        id = str(hash(datetime.datetime.now()))[1:7]
        from ..data import SingleLabelDataset
        if isinstance(train, SingleLabelDataset) and self.target != "single":
            print("You are using the model in multi mode but input is SingeleLabelDataset.")
            return 0

        validation=[]
        train_history = {"loss": []}

        assert not (type(train)== SingleLabelDataset and self.target=="multi"), "You inserted a SingleLabelDataset but chose multi as target."
        assert not (type(train) == MultiLabelDataset and self.target=="single"), "You inserted a MultiLabelDataset but chose single as target."


        best_loss = 10000000
        last_best_loss_update=0
        from ignite.metrics import Precision, Accuracy, Average
        for e in range(epochs):
            losses = {"loss": str(0.)}
            average = Average()
            train_loader = torch.utils.data.DataLoader(train, batch_size=batch_size, shuffle=True)

            with tqdm(train_loader,
                      postfix=[losses], desc="Epoch %i/%i" %(e+1,epochs), ncols=100) as pbar:
                for i, b in enumerate(train_loader):
                    self.optimizer.zero_grad()
                    y = b["labels"].to(self.device)

                    x = self.transform(b["text"]).to(self.device)
                    output = self(x)
                    if hasattr(self, "regularize"):
                        l = self.loss(output, y) + self.regularize()
                    else:
                        l = self.loss(output, y)
                    l.backward()

                    self.optimizer.step()
                    average.update(l.item())
                    pbar.postfix[0]["loss"] = round(average.compute().item(),2*self.PRECISION_DIGITS)
                    pbar.update()
                # torch.cuda.empty_cache()
                if valid is not None:
                    validation.append(self.evaluate_classes(classes_subset=classes_subset,
                                                           data=valid,
                                                           batch_size=valid_batch_size,
                                                           return_report= return_report,
                                                           return_roc=return_roc))
                    printable = validation[-1].copy()
                    if return_roc==True:
                        printable["auc_roc"] = (printable["auc_roc"][0], "...")
                    if return_report==True:
                        printable["report"] = (printable["report"]["micro avg"])

                    pbar.postfix[0].update(printable)
                    pbar.update()
            if patience > -1:
                if valid is None :
                    print("check validation loss")
                    if best_loss - average.compute().item() > tolerance:
                        print("update validation and checkoint")
                        best_loss = average.compute().item()
                        torch.save(self.state_dict(), id+"_checkpoint.pt")
                        #save states
                        last_best_loss_update = 0
                    else:
                        print("increment no epochs")
                        last_best_loss_update += 1

                    if last_best_loss_update >= patience:
                        print("breaking at %i" % (patience,))
                        print("Early Stopping.")
                        break
                elif valid is not None:
                    if best_loss - validation[-1]["valid_loss"] > tolerance:
                        best_loss = validation[-1]["valid_loss"]
                        torch.save(self.state_dict(), id+"_checkpoint.pt")
                        # save states
                        last_best_loss_update = 0
                    else:
                        last_best_loss_update += 1

                    if last_best_loss_update >= patience:
                        print("Early Stopping.")
                        break

            train_history["loss"].append(average.compute().item())
        if patience > -1:
            self.load_state_dict(torch.load(id+"_checkpoint.pt"))
        #Load best
        return{"train":train_history, "valid": validation }


    def predict(self, x, return_scores=False, tr=0.5, method="hard"):
        """
        Classify sentence string  or a list of strings.

        Args:
            x:  A list of the text instances.
            return_scores:  If True, the labels are returned with their corresponding confidence scores
            tr: The threshold at which the labels are returned.
            method: Method of thresholding
                    (hard will cutoff at ``tr``, mcut will look for the largest distance in
                    confidence between two labels following each other and will return everything above)

        Returns:
            A list of the labels

        """
        self.eval()
        print(self.training)
        if self.target =="single":
            method=="max"

        if not hasattr(self, "classes_rev"):
            self.classes_rev = {v: k for k, v in self.classes.items()}
        x = self.transform(x).to(self.device)
        with torch.no_grad(): output = self.act(self(x))
        prediction = self.threshold(output, tr=tr, method=method)
        self.train()
        if return_scores:
            return [[(self.classes_rev[i.item()], s[i].item()) for i in torch.where(p==1)[0]] for s, p in zip(output,prediction)]
        return [[self.classes_rev[i.item()] for i in torch.where(p==1)[0]] for p in prediction]

    def predict_dataset(self, data, batch_size=50, tr=0.5, method="hard"):
        """
        Predict all labels for a dataset int the mlmc.data.MultilabelDataset format.

        For detailed information on the arcuments see `mlmc.models.TextclassificationAbstract.predict`

        Args:
            data: A MultilabelDataset
            batch_size: Batch size
            tr: Threshold
            method: mcut or hard

        Returns:
            A list of labels

        """
        train_loader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=False)
        predictions = []
        for b in tqdm(train_loader, ncols=100):
            predictions.extend(self.predict(b["text"], tr=tr, method=method))
        return predictions

    def threshold(self, x, tr=0.5, method="hard"):
        """
        Thresholding function for outputs of the neural network.
        So far a hard threshold ( tr=0.5, method="hard")  is supported and
        dynamic cutting (method="mcut")

        This is wrapper for functions defined in py:mod:`mlmc.representations.output_transformations`

        Args:
            x: A tensor
            tr: Threshold
            method: mcut or hard

        Returns:

        """
        if method=="hard":
            return threshold_hard(x=x, tr=tr)
        if method=="mcut":
            return threshold_mcut(x)
        if method=="max":
            return threshold_max(x)

    def transform(self, x):
        """
        A standard transformation function from text to network input format

        The function looks for the tokenizer attribute. If it doesn't exist the transform function has to
        be implemented in the child class

        Args:
            x: A list of text

        Returns:
            A tensor in the network input format.

        """
        assert hasattr(self, 'tokenizer'), "If the model does not have a tokenizer attribute, please implement the" \
                                           "transform(self, x)  method yourself. TOkenizer can be allocated with " \
                                           "embedder, tokenizer = mlmc.helpers.get_embedding() or " \
                                           "embedder, tokenizer = mlmc.helpers.get_transformer()"
        return self.tokenizer(x,self.max_len)

    def _init_input_representations(self):
        if is_transformer(self.representation):
            if not hasattr(self, "n_layers"): self.n_layers=4
            try:
                if self.n_layers == 1:
                    self.embedding, self.tokenizer = get(model=self.representation)
                    self.embeddings_dim = self.embedding(torch.tensor([[0]]))[0].shape[-1]
                else:
                    self.embedding, self.tokenizer = get(model=self.representation, output_hidden_states=True)
                    self.embeddings_dim = \
                        torch.cat(self.embedding(self.embedding.dummy_inputs["input_ids"])[2][-self.n_layers:], -1).shape[-1]
            except TypeError:
                print("If your using a model that does not support returning hiddenstates, set n_layers=1")
                import sys
                sys.exit()
            for param in self.embedding.parameters(): param.requires_grad = False

        else:
            self.embedding, self.tokenizer = get(self.representation, freeze=True)
            self.embeddings_dim = self.embedding(torch.LongTensor([[0]])).shape[-1]
            for param in self.embedding.parameters(): param.requires_grad = False

    def num_params(self):
        """
        Count the number of trainable parameters.

        Returns:
            The number of trainable parameters
        """
        total = sum(p.numel() for p in self.parameters())
        trainable = sum(p.numel() for p in self.parameters() if p.requires_grad)
        print("Parameters:\n"
              "Trainable:\t%i\n"
              "Fixed:\t%i\n"
              "-----------\n"
              "Total:\t%i" % (trainable, total-trainable,total))


from mlmc.representation import get
import torch
from tqdm import tqdm

def get_word_embedding_mean(classes, model):
    emb, tok = get(model)
    transformed = tok(classes)
    mask = (transformed!=0).int()
    embeddings = emb(transformed)
    return embeddings.sum(-2)/mask.sum(-1, keepdim=True)

def get_lm_repeated(classes, model, repeat=20):
    emb, tok = get(model)
    transformed = tok([" ".join(cls.split()*repeat) for cls in classes])
    return emb(transformed)[1]

def get_lm_generated(classes, model, num=10, device="cuda:0"):
    with torch.no_grad():
        from transformers import AutoModelWithLMHead, AutoTokenizer
        gp2E = AutoModelWithLMHead.from_pretrained('gpt2').to(device)
        gpt2T= AutoTokenizer.from_pretrained("gpt2")

        generated_sentences = [[gpt2T.decode(x) for x in gp2E.generate(torch.tensor([gpt2T.encode((cls+" ")*3)]*num).to("cuda:0"), max_length=80)]
         for cls in tqdm(classes, total=len(classes))
         ]
        emb, tok = get(model)
        emb = emb.to(device)
        embeddings = [torch.cat([emb(tok(s).to(device))[1] for s in cls]).mean(0) for cls in generated_sentences]
    return torch.stack(embeddings,0)

def get_graph_augmented(classes, model, graph, topk=20, batch_size=64, device="cuda:0"):
    from ..graph.graph_loaders_elsevir import embed_align
    doct = embed_align(classes, graph=graph, model="glove300", topk=topk, batch_size=batch_size, device=device)

    emb, tok = get(model)
    emb.to(device)
    with torch.no_grad():
        transformed = tok([cls + " | " + ". ".join(keywords) for cls, keywords in doct.items()])
        embeddings = emb(transformed.to(device))[1]
    return embeddings



"""A collection of frequently used tranforming functions for strings or tensors"""

import torch

def clean(x):
    """
    Remove every character in a string that is not ascii, punctuation or whitespace
    :param x:
    :return:
    """
    import string
    return "".join([c for c in x if c in string.ascii_letters + string.punctuation + " "])

def label_smoothing(x, alpha=0.1):
    """
    Take one hot vector and return a smoothed version of it. ( [0,0,1] -> [0.05, 0.05, 0.9])
    :param x: one hot vector
    :param alpha: Amount of the smoothing
    :return: smoothed one hot vector
    """

    K = x.shape[-1]
    return (1 - alpha) * x + alpha / K

def label_smoothing_random(x, alpha=0.1):
    """
    Label smoothing with noise.
    :param x: One hot vector
    :param alpha: Amount of the smoothing
    :return: A smoothed one hot vector with added noise.
    """
    return (1 - alpha) * x + alpha * torch.rand(x.shape)



"""
A save and load function for models..
"""
from pathlib import Path

import torch
import warnings


def save(model, path, only_inference=False):
    """Saving a model to disk

    Args:
        model: The model
        path:  The path to save the model to
        only_inference: If False, the optimizer and loss function are saved to disk aswell.
    Returns:
        The path the model was saved to
    """
    # Remove  loss and optimizer from model, needs to be saved separately
    optimizer_tmp = model.optimizer
    loss_tmp = model.loss

    model.optimizer = None
    model.loss = None

    if only_inference:
        # Use torch.save to save the inference state. if save_all: Save the input representation (embedding or lm)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            torch.save(model, Path(path))

    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            args = type(model).__init__.__code__.co_varnames[1:-1] + tuple(set(super(type(model), model).__init__.__code__.co_varnames[1:-1])-set(["optimizer", "loss"]))
            values = {v: model.__dict__[v] for v in args if v != "kwargs"}
            torch.save({
                "type": type(model),
                "args": values,
                "optimizer": optimizer_tmp,
                "loss": loss_tmp,
                "model_state_dict": model.state_dict()}, path)


    # Reattach loss and optimizer and variables
    model.loss = loss_tmp
    model.optimizer = optimizer_tmp
    return path


def load(path, only_inference=False, only_cpu=False):
    """
    Load a model from disk

    Args:
        path: The path to load the model from
        only_inference: If the model was saved with only_inference=True this should be True aswell.
        only_cpu: If the model was trained on a GPU, but is trying to be loaded on a system without GPUs set ``only_cpu=True``
            to map the model into RAM instead of GPU.

    Returns:
        The loaded model.

    """
    additional_arguments = {}
    if only_cpu:
        additional_arguments["map_location"] = torch.device('cpu')

    if only_inference:
        loaded = torch.load(Path(path),**additional_arguments)
        return loaded
    else:
        #load all information
        loaded = torch.load(Path(path),**additional_arguments)
        # Create a model with the same parameters
        model = loaded["type"](**loaded["args"])
        model.load_state_dict(loaded["model_state_dict"])
        model.optimizer = loaded["optimizer"]
        model.loss = loaded["loss"]
        return model



"""A Collection of function to load, export and cache various datasets.
"""

import json
import pickle
import zipfile
from pathlib import Path
from urllib import error
from urllib.request import urlopen
from xml.etree import ElementTree
from zipfile import ZipFile

import networkx as nx
import tarfile
import tempfile
from bs4 import BeautifulSoup
from io import BytesIO
from sklearn.model_selection import train_test_split
from tqdm import tqdm

CACHE = Path.home() / ".mlmc" / "datasets"
URL = "https://aspra29.informatik.uni-leipzig.de:9090/"

def _load_from_tmp(dataset):
    if not Path.exists(CACHE):
        Path.mkdir(CACHE)
    if Path.is_file(CACHE / dataset):
        print("Loading from cache...")
        with open(CACHE / dataset, "rb") as f:
            data = pickle.load(f)
        return data
    else:
        return None

def _save_to_tmp(dataset, data):
    if not Path.exists(CACHE):
        Path.mkdir(CACHE)
    if not Path.is_file(CACHE / dataset):
        with open(CACHE / dataset, "wb") as f:
            pickle.dump(data, f)
        return True
    else:
        return False


# ============================================================================
# ============================================================================


def load_aapd():
    data = _load_from_tmp("aapd")
    if data is not None:
        return data
    else:
        try:
            resp = urlopen(URL + "AAPD.zip")
        except error.HTTPError:
            print(error.HTTPError)
            return None
        assert resp.getcode() == 200, "Download not found Error: (%i)" % (resp.getcode(),)
        zf = ZipFile(BytesIO(resp.read()))
        with zf.open("AAPD/aapd.json") as f:
            classes = json.load(f)
        data = {}
        for x in ("test", "val", "train"):
            with zf.open("AAPD/text_" + x) as f:
                test_x = [x.decode() for x in f.readlines()]
            with zf.open("AAPD/label_" + x) as f:
                test_y = [[y.lower() for y in x.decode().replace("\n", "").split(" ")] for x in f.readlines()]
            if x == "val":
                data["valid"] = [test_x, test_y]
            else:
                data[x] = [test_x, test_y]
        _save_to_tmp("aapd", (data, classes))
    return data, classes


def load_rcv1(path=None):
    data = _load_from_tmp("rcv1")
    if data is not None:
        return data

    assert path is not None, "RCV1 must be input with a path to the downloaded corpus, when first called"
    dataset = Path(path)
    with open(dataset / "categories.txt")as f:
        labels = {}
        needed_zips = set()
        for x in f.readlines():
            if x.split()[0][-4:] == ".zip" and len(x.replace("\n", "").split()[2:]) > 0:
                needed_zips.add(x.split()[0])
                labels[x.split()[1]] = x.replace("\n", "").split()[2:]

    with open(dataset / "train.split") as f:
        train_ids = [x.replace("\n", "") for x in f.readlines()]
    with open(dataset / "test.split") as f:
        test_ids = [x.replace("\n", "") for x in f.readlines()]

    with open(dataset / "topic_codes.txt") as f:
        classes = [(x.split("\t")[0], x.replace("\n", "").split("\t")[1:]) for x in f.readlines()[2:-1]]
        classes = set([x[0] for x in classes])

    with tempfile.TemporaryDirectory() as tempdir:
        for file in tqdm(needed_zips):
            zipfile.ZipFile(dataset / file).extractall(Path(tempdir))

        def _get(ids):
            errors = 0
            documents = []
            labelsets = []
            for nid in tqdm(ids):
                file = nid + "newsML.xml"
                try:
                    with open(Path(tempdir) / file) as tmp:
                        xml = ElementTree.fromstring(tmp.read())
                        text = ElementTree.tostring(xml, method='text').decode().replace("\n", " ").replace("  ",
                                                                                                            " ").strip()
                        documents.append(text)
                        labelsets.append([x for x in labels[file] if x in classes])
                except Exception:
                    try:
                        with open(Path(tempdir) / file, encoding="iso-8859-1") as tmp:
                            xml = ElementTree.fromstring(tmp.read())
                            text = ElementTree.tostring(xml, method='text').decode().replace("\n", " ").replace("  ",
                                                                                                                " ").strip()
                            documents.append(text)
                            labelsets.append([x for x in labels[file] if x in classes])
                    except Exception as e2:
                        errors += 1
                        print(e2)
            print(str(errors) + " Errors")
            return (documents, labelsets)

        train = _get(train_ids)
        test = _get(test_ids)

    classes = list(classes.intersection(set([x for y in train[1] + test[1] for x in y])))
    classes = dict(zip(classes, range(len(classes))))
    data = {"train": train, "test": test}

    with open(dataset / "topic_codes.txt", "r") as f:
        topics = [x.replace("\n", "").split("\t") for x in f.readlines() if len(x) > 1][2:]
    topicmap = {x[0]: x[1] for x in topics}
    data["topicmap"] = topicmap

    with open(dataset / "rcv1.topics.hier.orig", "r") as f:
        content = f.readlines()
    import re
    edges = [(re.split(" +", x)[1], re.split(" +", x)[3]) for x in content]
    edges = [(topicmap.get(x[0],x[0]).capitalize(),topicmap.get(x[1],x[1]).capitalize()) for x in edges]
    graph = nx.DiGraph(edges[1:])
    data["graph"] = graph


    for key in ("train", "test"):
        data[key] = (data[key][0],[[topicmap[l].capitalize() for l in labellist] for labellist in data[key][1]])
    classes = {topicmap[k].capitalize(): v for k, v in classes.items()}

    data["topicmap"] = {v.capitalize():k for k,v in topicmap.items()}

    _save_to_tmp("rcv1", (data, classes))
    return data, classes


def load_wiki30k(path="/disk1/users/jborst/Data/Test/MultiLabel/wiki30k"):
    import pickle
    with open(Path(path) / "wiki30k_raw_text.p", "rb") as f:
        content = pickle.load(f)
    train_x = [x["text"] for x in content[0]]
    train_y = [x["catgy"] for x in content[0]]
    test_x = [x["text"] for x in content[1]]
    test_y = [x["catgy"] for x in content[1]]
    data = {"train": (train_x, train_y), "test": (test_x, test_y)}
    return data, content[3]


def load_eurlex():
    data = _load_from_tmp("eurlex")
    if data is not None:
        return data
    else:
        try:
            resp = urlopen(URL + "EURLex.zip")
        except error.HTTPError:
            print(error.HTTPError)
            return None
        assert resp.getcode() == 200, "Download not found Error: (%i)" % (resp.getcode(),)
        zf = ZipFile(BytesIO(resp.read()))
        with zf.open("EURLex/eurlex_raw_text.p") as f:
            content = pickle.load(f)
        train_x = [x["text"] for x in content[0]]
        train_y = [x["catgy"] for x in content[0]]
        test_x = [x["text"] for x in content[1]]
        test_y = [x["catgy"] for x in content[1]]
        classes = content[3]

        reverse_classes = {v: k for k, v in classes.items()}
        train_y = [[reverse_classes[x] for x in labels] for labels in train_y]
        test_y = [[reverse_classes[x] for x in labels] for labels in test_y]
        data = {}
        data["train"] = (train_x, train_y)
        data["test"] = (test_x, test_y)

        _save_to_tmp("eurlex", (data, classes))
    return data, classes


def load_huffpost(test_split=0.25):
    data = _load_from_tmp("huffpost")
    if data is not None:
        return data
    else:
        try:
            resp = urlopen(URL + "HuffPost.zip")
        except error.HTTPError:
            print(error.HTTPError)
            return None
        assert resp.getcode() == 200, "Download not found Error: (%i)" % (resp.getcode(),)
        zipfile = ZipFile(BytesIO(resp.read()))

        with zipfile.open("HuffPost/News_Category_Dataset_v2.json", "r") as f:
            content = json.loads("[" + ",".join([x.decode() for x in f.readlines()]) + "]")
        headlines = [x["headline"] for x in content]
        label = [x["category"] for x in content]
        data = (headlines, label)
        classes = list(set(label))
        classes = dict(zip(classes, range(len(classes))))
        n_arr = len(data)
        data = train_test_split(*data, test_size=test_split)
        tmp = [[], []]
        for i, arr in enumerate(data):
            tmp[i % n_arr].append(arr)

        _save_to_tmp("huffpost", (data, classes))

    return {"train": tmp[0], "test": tmp[1]}, classes


def load_moviesummaries(test_split=0.25):
    data = _load_from_tmp("moviesummaries")
    if data is not None:
        return data
    else:

        try:
            resp = urlopen(URL + "MovieSummaries.zip")
        except error.HTTPError:
            print(error.HTTPError)
            return None
        assert resp.getcode() == 200, "Download not found Error: (%i)" % (resp.getcode(),)
        zf = ZipFile(BytesIO(resp.read()))

        from sklearn.model_selection import train_test_split
        with zf.open("MovieSummaries/plot_summaries.txt", "r") as f:
            content = [x.decode() for x in f.readlines()]
        with zf.open("MovieSummaries/movie.metadata.tsv", "r") as f:
            meta = {x.decode().split("\t")[0]: x.decode().split("\t")[-1].replace("\n", "") for x in
                    f.readlines()}  # f.readlines()#[x.split("\t")[-1].replace("\n","") for x in f.readlines()]
        meta = {k: [x.split(": ")[-1].replace("\"", "").replace("}", "") for x in genre.split(", ")] for k, genre in
                meta.items()}
        data = [(x.split("\t")[1], meta[str(x.split("\t")[0])]) for x in content if
                str(x.split("\t")[0]) in meta.keys()]
        text = [x[0] for x in data]
        label = [x[1] for x in data]

        data = (text, label)
        classes = list(set([x for y in label for x in y]))
        classes = dict(zip(classes, range(len(classes))))

        n_arr = len(data)
        data = train_test_split(*data, test_size=test_split)
        tmp = [[], []]
        for i, arr in enumerate(data):
            tmp[i % n_arr].append(arr)

        dataset=  {"train": tmp[0], "test": tmp[1]}
        _save_to_tmp("moviesummaries", (dataset, classes))

        return dataset, classes


# ----------------------------------------------
# Hierarchical Multilabel
# see http://kt.ijs.si/DragiKocev/PhD/resources/doku.php?id=hmc_classification
# ----------------------------------------------

def load_blurbgenrecollection():
    url = "https://fiona.uni-hamburg.de/ca89b3cf/blurbgenrecollectionen.zip"
    data = _load_from_tmp("blurbgenrecollection")
    if data is not None:
        return data
    else:
        resp = urlopen(url)
        zf = ZipFile(BytesIO(resp.read()))
        data = {}
        for purpose in ["dev", "train", "test"]:
            soup = BeautifulSoup(
                "<root>" + zf.open("BlurbGenreCollection_EN_" + purpose + ".txt").read().decode("utf-8").replace(
                    "\n", "") + "</root>")
            text, labels = [], []
            for i in soup.findAll("book"):
                text.append(i.find("body").text)
                labels.append([x.text for x in i.find("topics").children])
            if purpose == "dev":
                data["valid"] = (text, labels)
            else:
                data[purpose] = (text, labels)

        edges = [x.decode("utf-8").replace("\n", "").split("\t") for x in zf.open("hierarchy.txt").readlines()]
        classes = list(set([x for y in data["train"][1] + data["valid"][1] + data["test"][1] + edges for x in y]))
        classes = dict(zip(classes, range(len(classes))))
        edges = [e for e in edges if len(e) == 2]
        graph = nx.DiGraph()
        graph.add_nodes_from(classes.keys())
        graph.add_edges_from(edges)
        data["graph"] = graph
        _save_to_tmp("blurbgenrecollection", (data, classes))
        return data, classes


def load_blurbgenrecollection_de():
    url = "https://www.inf.uni-hamburg.de/en/inst/ab/lt/resources/data/germeval-2019-hmc/germeval2019t1-public-data-final.zip"
    data = _load_from_tmp("blurbgenrecollection_de")
    if data is not None:
        return data
    else:

        resp = urlopen(url)
        zipfile = ZipFile(BytesIO(resp.read()))

        data = {}
        for purpose in ["dev", "train", "test"]:
            soup = BeautifulSoup(
                "<root>" + zipfile.open("blurbs_" + purpose + ".txt").read().decode(
                    "utf-8").replace("\n", "") + "</root>")
            text, labels = [], []
            for i in soup.findAll("book"):
                text.append(i.find("body").text)
                labels.append([x.text for x in i.find("categories").findAll("topic")])
            labels = [list(set(x)) for x in labels]
            if purpose == "dev":
                data["valid"] = (text, labels)
            else:
                data[purpose] = (text, labels)

        edges = [x.decode("utf-8").replace("\n", "").split("\t") for x in zipfile.open("hierarchy.txt").readlines()]
        classes = list(set([x for y in data["train"][1] + data["valid"][1] + data["test"][1] + edges for x in y]))
        classes = dict(zip(classes, range(len(classes))))
        edges = [e for e in edges if len(e) == 2]
        graph = nx.DiGraph(edges)
        data["graph"] = graph
        _save_to_tmp("blurbgenrecollection_de", (data, classes))
        return data, classes


def load_webofscience():
    raise NotImplementedError
    # url = "https://data.mendeley.com/datasets/9rw3vkcfy4/6/files/c9ea673d-5542-44c0-ab7b-f1311f7d61df/WebOfScience.zip?dl=1"
    # data = _load_from_tmp("blurbgenrecollection")
    # if data is not None:
    #     return data
    # else:
    #     from xml.etree import ElementTree
    #     from urllib.request import urlopen
    #     from zipfile import ZipFile
    #     from io import BytesIO
    #     import re
    #     resp = urlopen(url)
    #     zipfile = ZipFile(BytesIO(resp.read()))

def load_20newsgroup():
    url = "http://qwone.com/~jason/20Newsgroups/20news-bydate.tar.gz"
    data = _load_from_tmp("20newsgroup")
    if data is not None:
        return data
    else:

        with tempfile.TemporaryDirectory() as tmpdir:

            resp = urlopen(url)
            tf = tarfile.open(fileobj=resp, mode="r|gz")
            tf.extractall(Path(tmpdir))
            testdir = Path(tmpdir) / '20news-bydate-test'
            traindir = Path(tmpdir) / '20news-bydate-train'

            classes = []

            text, label = [], []
            for catg in testdir.iterdir():
                classes.append(catg.name)
                for file in catg.iterdir():
                    with open(file, 'r', encoding="ISO-8859-1") as f:
                        text.append(f.read())
                        label.append([catg.name])
            testdata = (text, label)

            text, label = [], []
            for catg in traindir.iterdir():
                classes.append(catg.name)
                for file in catg.iterdir():
                    with open(file, 'r', encoding="ISO-8859-1") as f:
                        text.append(f.read())
                        label.append([catg.name])
            traindata = (text, label)

            classes = list(set(classes))
            classes.sort()
            classes = dict(zip(classes, range(len(classes))))

            edges = [
                (x.split(".")[y], x.split(".")[y + 1] if y + 2 != len(x.split(".")) else x)
                for x in classes for y in range(len(x.split(".")) - 1)]
            edges += list(set([("ROOT", x.split(".")[0]) for x in classes]))

            graph = nx.DiGraph(edges)
            data = {"train": traindata, "test": testdata, "graph": graph}
        _save_to_tmp("20newsgroup", (data, classes))

        return data, classes

def load_agnews():
    url = "https://s3.amazonaws.com/fast-ai-nlp/ag_news_csv.tgz"
    data = _load_from_tmp("agnews")
    if data is not None:
        return data
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            resp = urlopen(url)
            tf = tarfile.open(fileobj=resp, mode="r|gz")
            tf.extractall(Path(tmpdir))
            testdir = Path(tmpdir) / 'ag_news_csv/test.csv'
            traindir = Path(tmpdir) / 'ag_news_csv/train.csv'
            classesdir = Path(tmpdir) / 'ag_news_csv/classes.txt'

            with open(testdir, "r") as f:
                testdata = [x.replace("\n", "").split('","')[::-1] for x in f.readlines()]
                testlabel = [int(x[2].replace('"', '')) for x in testdata]
                testtitle = [x[1] for x in testdata]
                testdescription = [x[0] for x in testdata]
                testtext = [" \n ".join([t, d]) for t, d in zip(testtitle, testdescription)]
            with open(traindir, "r") as f:
                traindata = [x.replace("\n", "").split('","')[::-1] for x in f.readlines()]
                trainlabel = [int(x[2].replace('"', '')) for x in traindata]
                traintitle = [x[1] for x in traindata]
                traindescription = [x[0] for x in traindata]
                traintext = [" \n ".join([t, d]) for t, d in zip(traintitle, traindescription)]
            with open(classesdir,"r") as f:
                classes = [x.replace("\n","") for x in f.readlines()]
                classes = dict(zip(classes, range(len(classes))))
                rev_classes = {v: k for k, v in classes.items()}
        data = {
            "train": (traintext,[[rev_classes[x-1]] for x in trainlabel]),
            "test": (testtext, [[rev_classes[x-1]] for x in testlabel]),
            "test_title": testtitle,
            "test_description": testdescription,
            "train_title": traintitle,
            "train_description": traindescription
        }
        _save_to_tmp("agnews", (data, classes))
        return data, classes

def load_dbpedia():
    url = "https://github.com/le-scientifique/torchDatasets/raw/master/dbpedia_csv.tar.gz"
    data = _load_from_tmp("dbpedia")
    if data is not None:
        return data
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            resp = urlopen(url)
            tf = tarfile.open(fileobj=resp, mode="r|gz")
            tf.extractall(Path(tmpdir))
            testdir = Path(tmpdir) / 'dbpedia_csv/test.csv'
            traindir = Path(tmpdir) / 'dbpedia_csv/train.csv'
            classesdir = Path(tmpdir) / 'dbpedia_csv/classes.txt'
#
            with open(testdir, "r", encoding="iso-8859-1") as f:
                testdata = [x.replace("\n", "").split(',') for x in f.readlines()]
                testlabel = [int(x[0]) for x in testdata]
                testtitle = [x[1] for x in testdata]
                testdescription = [x[2] for x in testdata]
                testtext = [" \n ".join([t, d]) for t, d in zip(testtitle, testdescription)]
            with open(traindir, "r", encoding="iso-8859-1") as f:
                traindata = [x.replace("\n", "").split(',') for x in f.readlines()]
                trainlabel = [int(x[0]) for x in traindata]
                traintitle = [x[1] for x in traindata]
                traindescription = [x[2] for x in traindata]
                traintext = [" \n ".join([t, d]) for t, d in zip(traintitle, traindescription)]
            with open(classesdir,"r") as f:
                classes = [x.replace("\n","") for x in f.readlines()]
                classes = dict(zip(classes, range(len(classes))))
                rev_classes = {v: k for k, v in classes.items()}
        data = {
            "train": (traintext, [[rev_classes[x-1]] for x in trainlabel]),
            "test": (testtext, [[rev_classes[x-1]] for x in testlabel]),
            "test_title": testtitle,
            "test_description": testdescription,
            "train_title": traintitle,
            "train_description": traindescription
        }
        _save_to_tmp("dbpedia", (data, classes))
        return data, classes

def load_ohsumed():
    url = "http://disi.unitn.eu/moschitti/corpora/ohsumed-first-20000-docs.tar.gz"
    url_classes = "http://disi.unitn.eu/moschitti/corpora/First-Level-Categories-of-Cardiovascular-Disease.txt"
    data = _load_from_tmp("ohsumed")
    if data is not None:
        return data
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            resp = urlopen(url)
            tf = tarfile.open(fileobj=resp, mode="r|gz")
            tf.extractall(Path(tmpdir))
            testdir = Path(tmpdir) / 'ohsumed-first-20000-docs/test'
            traindir = Path(tmpdir) / 'ohsumed-first-20000-docs/training'

            testdict = {}
            for catg in testdir.iterdir():
                catg_name = int(catg.name[1:].replace("0", ""))
                for file in catg.iterdir():
                    if file.name not in testdict:
                        testdict[file.name] = [catg_name]
                    else:
                        if int(catg_name) not in testdict[file.name]:
                            testdict[file.name].append(catg_name)

            traindict = {}
            for catg in traindir.iterdir():
                catg_name = int(catg.name[1:].replace("0", ""))
                for file in catg.iterdir():
                    if file.name not in traindict:
                        traindict[file.name] = [catg_name]
                    else:
                        if int(catg_name) not in traindict[file.name]:
                            traindict[file.name].append(catg_name)

            testdata, testlabel, testlist = [], [], []
            for catg in testdir.iterdir():
                for file in catg.iterdir():
                    if file.name not in testlist:
                        testlist.append(file.name)
                        with open(file, 'r') as f:
                            testdata.append(f.read().split(("\n"), 1))
                            testlabel.append(testdict.get(file.name))
            testtitle = [x[0] for x in testdata]
            testdescription = [x[1].replace("\n", "").strip() for x in testdata]
            testtext = [" \n ".join([t, d]) for t, d in zip(testtitle, testdescription)]

            traindata, trainlabel, trainlist = [], [], []
            for catg in traindir.iterdir():
                for file in catg.iterdir():
                    if file.name not in trainlist:
                        trainlist.append(file.name)
                        with open(file, 'r') as f:
                            traindata.append(f.read().split(("\n"), 1))
                            trainlabel.append(traindict.get(file.name))
            traintitle = [x[0] for x in traindata]
            traindescription = [x[1].replace("\n", "").strip() for x in traindata]
            traintext = [" \n ".join([t, d]) for t, d in zip(traintitle, traindescription)]

            classes_file = urlopen(url_classes).read().decode("utf-8")
            classes_file = classes_file.split("\n")
            classes_list = [x[:-3].strip() for x in classes_file]
            classes_list.pop()
            classes = dict(zip(classes_list, range(len(classes_list))))
            rev_classes = {v: k for k, v in classes.items()}

            testlabellist, trainlabellist, tmp = [], [], []
            for x in testlabel:
                for v in x:
                    tmp.append(rev_classes[v-1])
                testlabellist.append(tmp)
                tmp = []

            for x in trainlabel:
                for v in x:
                    tmp.append(rev_classes[v-1])
                trainlabellist.append(tmp)
                tmp = []
        data = {
            "train": (traintext, trainlabellist),
            "test": (testtext, testlabellist),
            "test_title": testtitle,
            "test_description": testdescription,
            "train_title": traintitle,
            "train_description": traindescription
        }
        _save_to_tmp("ohsumed", (data, classes))
        return data, classes

def export(data, classes, path=Path("./export")):
    path = Path(path)
    if not path.exists():
        path.mkdir()
    for k, v in data.items():
        if k in ("test", "valid", "train"):
            with open(path / k + "_x.txt", "w") as o:
                o.writelines([x.replace("\n", "\t") + "\n" for x in v[0]])
            with open(path / k + "_y.txt", "w") as o:
                o.writelines(["\t".join(x) + "\n" for x in v[1]])
        elif k == "graph":
            edgelist = [t[0] + "\t" + t[1] + "\n" for t in v.edges]
            with open(path / "edge_list.txt", "w") as o:
                o.writelines(edgelist)

    with open(path / "classes.txt", "w") as o:
        o.writelines([x + "\n" for x in classes.keys()])



"""
Loading Embeddings and Word embeddings in an automated fashion.
"""
from pathlib import Path
import shutil
from urllib import error
from urllib.request import urlopen
from zipfile import ZipFile

import numpy as np
import os
import torch
from io import BytesIO
from transformers import *

dir_path = Path(os.path.dirname(os.path.realpath(__file__)))
with open(dir_path/"model.txt", "r") as f:
    MODELS = {k.replace("\n", ""): (AutoModel, AutoTokenizer, k.replace("\n","")) for k in f.readlines()}

for k, v in {"bert": (BertModel, BertTokenizer, 'bert-large-uncased'),
             "albert": (AlbertModel, AlbertTokenizer, 'albert-large-v2'),
             "ctrl": (CTRLModel, CTRLTokenizer, 'ctrl'),
             "distilbert": (DistilBertModel, DistilBertTokenizer, 'distilbert-base-uncased'),
             "roberta": (RobertaModel, RobertaTokenizer, 'roberta-base'),
             }.items():
    if k not in MODELS.keys():
        MODELS[k]=v

STATICS = {
    "glove50": "glove.6B.50d.txt",
    "glove100": "glove.6B.100d.txt",
    "glove200": "glove.6B.200d.txt",
    "glove300": "glove.6B.300d.txt"
}

EMBEDDINGCACHE = Path.home() / ".mlmc" / "embedding"
EMBEDDINGCACHEINDEX = Path.home() / ".mlmc" / "embedding" / "index.txt"

def custom_embedding(name, file):
    """
    Add a custom static embedding file to the cache.

    Use this to register a new embedding file. The file should be a txt file and in the glove format. After loading the file
    with this function you will be able to load the embedding in any model by its name.

    Args:
        name: Name of the embedding for later use
        file: File of the embedding
    Returns:

    """
    source_location = Path(file)
    target_location = EMBEDDINGCACHE / source_location.name
    if (EMBEDDINGCACHE/source_location.name).exists():
        print("Embedding already cached")
    else:
        shutil.copy(source_location, target_location)
        with open(EMBEDDINGCACHEINDEX, "a") as f:
            f.write(name + "\t" + source_location.name)
        print("Cached successfully you can now load [%s] in any model." % (name,))

def delete_custom_embeddings():
    """
    Delete Custom Embeddings from Cache
    Returns:

    """
    with open (EMBEDDINGCACHEINDEX, "r") as f:
        for x in f.read().split("\n"):
            fp = EMBEDDINGCACHE / x.split("\t")[1]
            print(fp)
            fp.unlink()
    EMBEDDINGCACHEINDEX.unlink()

def empty_cache():
    """
    Delete everything from representations cache.

    This does only include static embeddings and custom embeddings. Language Models are downloaded in a different location
    defined by the huggingface library for now.

    ToDo: Use mlmc cache for huggingface to have more control when trying to free disk space.

    Returns:

    """
    for file in  EMBEDDINGCACHE.iterdir():
        file.unlink()

def add_test_example():
    if not EMBEDDINGCACHE.exists():
        EMBEDDINGCACHE.mkdir(parents=True)
    custom_embedding("test", dir_path / "custom_embedding.txt")


def load_static(embedding):
    """
    Load the embedding from a testfile.

    Args:
        embedding: one of [glove50, glove100, glove200, glove300]

    Returns: The embedding matrix and the vocabulary.

    """
    if EMBEDDINGCACHEINDEX.exists():
        with open (EMBEDDINGCACHEINDEX, "r") as f:
            for x in f.read().split("\n"):
                STATICS[x.split("\t")[0]] = x.split("\t")[1]

    if not (EMBEDDINGCACHE / STATICS[embedding]).exists():
        try:
            resp = urlopen("http://nlp.stanford.edu/data/glove.6B.zip")
        except error.HTTPError:
            print(error.HTTPError)
            return None
        assert resp.getcode() == 200, "Download not found Error: (%i)" % (resp.getcode(),)
        print("Downloading glove vectors... This may take a while...")
        zipfile = ZipFile(BytesIO(resp.read()))
        zipfile.extractall(EMBEDDINGCACHE)
    fp = EMBEDDINGCACHE / STATICS[embedding]

    glove = np.loadtxt(fp, dtype='str', comments=None)
    glove = glove[np.unique(glove[:, :1], axis=0, return_index=True)[1]]
    words = glove[:, 0]
    weights = glove[:, 1:].astype('float')
    weights = np.vstack((
        np.array([0] * len(weights[1])),  # the vector for the masking
        weights,
        np.mean(weights, axis=0)),  # the vector for the masking
    )
    words = words.tolist()+["<UNK_TOKEN>"]
    vocabulary = dict(zip(words, range(1, len(words) + 1)))
    return weights, vocabulary


def map_vocab(query, vocab, maxlen):
    """
    Map a query ( a list of lists of tokens ) to indices using the vocab mapping and
    pad (or cut respectively) all to maxlen.
    Args:
        query: a list of lists of tokens
        vocab: A mapping from tokens to indices
        maxlen: Maximum lengths of the lists

    Returns: A torch.Tensor with shape (len(query), maxlen)

    """
    ind = [[vocab.get(token, vocab["<UNK_TOKEN>"]) for token in s] for s in query]
    result = torch.zeros((len(query), maxlen)).long()
    for i, e in enumerate(ind):
       result[i,:min(len(e), maxlen)] = torch.LongTensor(e[:min(len(e), maxlen)])
    return result

def get_white_space_tokenizer(v):
    def tokenizer(x, maxlen=500):
        x = [x] if isinstance(x, str) else x
        x = [s.lower().split() for s in x]
        return map_vocab(x, v, maxlen).long()
    return tokenizer

def get_embedding(name, **kwargs):
    """
    Load a static word embedding from file.
    Args:
        name: File name of the word embedding. (Expects a text file in the glove format)
    Returns: A tuple of embedding and corresponding tokenizer
    """
    import h5py

    file = EMBEDDINGCACHE / (name + ".h5")
    if not (file).exists():
        weights, vocabulary = load_static(name)
        with h5py.File(file, "w") as f:
            f.create_dataset("weights", data=weights)
            f.create_dataset("vocab", data=str(vocabulary))
    else:
        with h5py.File(file, "r")as f:
            weights = f["weights"][()]
            vocabulary = eval(f["vocab"][()])


    e = torch.nn.Embedding(weights.shape[0], weights.shape[1],)
    e = e.from_pretrained(torch.Tensor(weights).float(), **kwargs)

    tokenizer = get_white_space_tokenizer(vocabulary)
    return e, tokenizer


def get_transformer(model="bert", **kwargs):
    """
    Get function for transformer models
    Args:
        model: Model name
        **kwargs: Additional keyword arguments

    Returns:  A tuple of embedding and corresponding tokenizer

    """
    model_class, tokenizer_class, pretrained_weights = MODELS.get(model,(None,None,None))
    if model_class is None:
        print("Model is not a transformer...")
        return None
    else:
        # Load pretrained model/tokenizer
        tokenizer = tokenizer_class.from_pretrained(pretrained_weights)

        def list_tokenizer(x, maxlen=500, return_start=False):
            x = [x] if isinstance(x, str) else x
            if return_start:
                i = [tokenizer.tokenize(sentence, add_special_tokens=False, pad_to_max_length=True) for sentence in x]
                ind = [torch.tensor([x.startswith("Ġ") or i == 0 for i, x in enumerate(sentence)]) for sentence in i]
                i = torch.nn.utils.rnn.pad_sequence([torch.tensor(tokenizer.convert_tokens_to_ids(a)) for a in i],
                                                    batch_first=True, padding_value=tokenizer.pad_token_id)
                return i, torch.nn.utils.rnn.pad_sequence(ind, batch_first=True, padding_value=False)
            else:
                i = torch.nn.utils.rnn.pad_sequence(
                    [torch.tensor([tokenizer.encode(sentence, add_special_tokens=False, pad_to_max_length=True)][0]) for
                     sentence in x], batch_first=True, padding_value=tokenizer.pad_token_id)
            i = i[:, :min(maxlen, i.shape[-1])]
            return i

        model = model_class.from_pretrained(pretrained_weights, **kwargs)
        return model, list_tokenizer

def get(model, **kwargs):
    """
        Universal get function for text embedding methods and tokenizers.

    Args:
        model: Model name (one of [ glove50, glove100, glove200, glove300] or any of the models on https://huggingface.co/models
        **kwargs:  Additional arguments in case of transformers. for example ``output_hidden_states=True`` for returning hidden states of transformer models.
            For details on the parameters for the specific models see https://huggingface.co


    Returns:
         A tuple of embedding and corresponding tokenizer

    Examples:
        ```
        embedder, tokenizer = get("bert-base-uncased")
        embedding = embedder(tokenizer("A sentence of various words"))
        ```
    The variable ``embedding`` will contain a torch tensor of shape

     (1, sequence_length, embedding_dim)


    """
    logging.getLogger("transformers.tokenization_utils").setLevel(logging.ERROR)
    module = get_transformer(model, **kwargs)
    if module is None:
        module = get_embedding(model, **kwargs)
        if module is None:
            raise FileNotFoundError
        return module
    else:
        return module

def is_transformer(name):
    """
    A check function. True if the ``name`` argument if found to be a valid transformer model name.

    Args:
        name: model name (see get)

    Returns: bool

    """
    return name in MODELS.keys()


"""
Multi-Label Zero-Shot Learning with Structured Knowledge Graphs Lee, Fang, Yeh (2018)
"""

import torch
from ..models.abstracts import TextClassificationAbstract
from ..layers import  LabelEmbeddingScoring
from ..representation.representations import get
import torch_geometric as torch_g


class SKG(TextClassificationAbstract):
    def __init__(self, adjacency, label_embed, classes, static=None, transformer="bert", max_len=300, **kwargs):
        super(SKG,self).__init__(**kwargs)
        self.classes=classes
        self.n_classes = len(classes)
        self.adjacency = torch.nn.Parameter(torch.from_numpy(adjacency).long(),requires_grad=False)
        self.edge_list = torch.nn.Parameter(torch.stack(torch.where(self.adjacency == 1), dim=0).long(),requires_grad=False)
        self.max_len = max_len
        self.static=static
        self.transformer=transformer

        self.embedder, self.tokenizer = get(static, transformer, output_hidden_states=True)
        self.embedder.eval()
        self.embedding_dim = torch.cat(self.embedder(self.embedder.dummy_inputs["input_ids"])[2][-5:-1],-1).shape[-1]
        self.scorer = LabelEmbeddingScoring(self.n_classes,
                                            self.embedding_dim,
                                            label_embed,
                                            label_freeze=True)

        self.ggc = torch_g.nn.GatedGraphConv(self.n_classes, 10)
        self.build()

    def forward(self, x):
        if self.n_layers == 1:
            with torch.no_grad():
                e = self.embedder(x)[0].permute(0, 2, 1)
        else:
            with torch.no_grad():
                e = torch.cat(self.embedder(x)[2][(-1 - self.n_layers):-1], -1).permute(0, 2, 1)

        initial_belief = self.scorer(e).sum(-2)
        updated_belief = self.ggc(initial_belief, edge_index=self.adjacency)
        return updated_belief



"""
Few-Shot and Zero-Shot Multi-Label Learning for Structured Label Spaces - Rios & Kavuluru (2018)
"""
import torch
from ..models.abstracts import TextClassificationAbstract
from ..representation import get, is_transformer
import re

class ZAGCNN(TextClassificationAbstract):
    def __init__(self, classes,   adjacency, representation="roberta", max_len=300, dropout = 0.5, **kwargs):
        super(ZAGCNN, self).__init__(**kwargs)

        self.classes = classes
        self.n_classes = len(classes)
        self.max_len = max_len
        self.use_dropout = dropout
        self.filters = 300
        self.hidden_dim = 300
        self.kernel_sizes = [3,10]
        self.adjacency=adjacency
        self.dropout = dropout
        self.adjacency_parameter = torch.nn.Parameter(torch.from_numpy(adjacency).float())
        self.adjacency_parameter.requires_grad = False

        assert not is_transformer(representation), "This model does not support language Models. See ZAGCNNLM."
        self.representation = representation
        self._init_input_representations()
        self.embedding.requires_grad=False
        self.label_embeddings = torch.nn.Parameter(
            self.embedding(self.tokenizer(
                [" ".join(re.split("[/ _-]", x.lower())) for x in self.classes.keys()]

            )[:, :4]).mean(-2)
        )
        self.label_embeddings.requires_grad = False


        self.convs = torch.nn.ModuleList(
            [torch.nn.Conv1d(self.embeddings_dim, self.filters, k) for k in self.kernel_sizes])
        self.pool = torch.nn.MaxPool1d(3, stride=2)
        self.document_projection = torch.nn.Linear(self.filters, self.embeddings_dim).to(self.device)



        self.dropout_layer= torch.nn.Dropout(self.dropout)
        import torch_geometric as torchg
        self.gcn1 = torchg.nn.GCNConv(in_channels=self.embeddings_dim, out_channels=self.hidden_dim)
        self.gcn2 = torchg.nn.GCNConv(in_channels=self.hidden_dim, out_channels=self.hidden_dim)

        self.projection = torch.nn.Linear(in_features=self.embeddings_dim, out_features=self.hidden_dim + self.label_embeddings.shape[-1])
        self.build()

    def forward(self, x):
        if self.n_layers == 1:
            with torch.no_grad():
                embedded = self.embedding(x)[0].permute(0, 2, 1)
        else:
            with torch.no_grad():
                embedded = torch.cat(self.embedding(x)[2][-self.n_layers:-1], -1).permute(0, 2, 1)
        embedded = self.dropout_layer(embedded)
        c = torch.cat([self.pool(torch.nn.functional.relu(conv(embedded.permute(0,2,1)))) for conv in self.convs], dim=-1).permute(0,2,1)
        d2 = torch.tanh(self.document_projection(c))
        a = torch.softmax(torch.matmul(d2, self.label_embeddings.t()),-1)

        label_wise_representation = torch.matmul(a.permute(0,2,1), c)

        label_wise_representation = self.dropout_layer(label_wise_representation)

        labelgcn = self.gcn1(self.label_embeddings, torch.stack(torch.where(self.adjacency_parameter == 1), dim=0))
        labelgcn = self.dropout_layer(labelgcn)
        labelgcn = self.gcn2(labelgcn, torch.stack(torch.where(self.adjacency_parameter == 1), dim=0))
        labelvectors = torch.cat([self.label_embeddings, labelgcn], dim=-1)
        return (torch.relu(self.projection(label_wise_representation))*labelvectors).sum(-1)

    def create_labels(self, classes):
        self.classes = classes
        del self.label_embeddings
        self.label_embeddings = torch.nn.Parameter(
            self.embedding(self.tokenizer([" ".join(re.split("[/ _-]", x.lower()) ) for x in self.classes.keys()])[:, :4].to(self.device)).mean(-2)
        )
        self.label_embeddings.requires_grad = False

    def set_adjacency(self, adj):
        del self.adjacency_parameter
        self.adjacency_parameter = torch.nn.Parameter(torch.from_numpy(adj).float().to(self.device))
        self.adjacency_parameter.requires_grad = False


import torch
from .abstracts import TextClassificationAbstract
from ..representation import get
##############################################################################################
##############################################################################################
#  Implementations
##############################################################################################


class KimCNN(TextClassificationAbstract):
    """
    Implementation of Yoon Kim 2014 KimCNN Classification network for Multilabel Application (added support for Language Models).
    """
    def __init__(self, classes, mode="transformer", representation="roberta", kernel_sizes=[3,4,5,6], filters=100, dropout=0.5, n_layers=4,max_len=200, **kwargs):
        """Class constructor and intialization of every hyperparameters

        :param classes:  A dictionary of the class label and the corresponding index
        :param mode:  One of (trainable, untrainable, multichannel, transformer).
                    Transformer has to be used in combination with representation being a transformer model name (see: https://huggingface.co/transformers/pretrained_models.html).
                    In combination with the other three one of the glove embeddings can be used (glove50, glove100, glove200, glove300)
                    'trainable' will finetune the wordembedding used, whereas "untrainable" will freeze the embedding layer.
                    'multichannel' will combine two embedding layers, one for finetuning on the task, one frozen.
        :param representation: The name of the representation to use. glove* or one of the hugginface transformers models.
        :param kernel_sizes: Sizes of the kernel used for the convolution
        :param filters: Number of filters used in the convolution
        :param dropout: Droupout rate
        :param max_len: Maximum length input sequences. Longer sequences will be cut.
        :param kwargs: Optimizer and loss function keyword arguments, see `mlmc.models.TextclassificationAbstract`
         """
        super(KimCNN, self).__init__(**kwargs)

        self.classes = classes
        self.n_classes = len(classes)
        self.max_len = max_len
        self.modes = ("trainable","untrainable","multichannel","transformer")
        self.mode = mode
        self.l = 1
        self.kernel_sizes = kernel_sizes
        self.filters = filters
        self.representation = representation
        self.dropout = dropout
        self.n_layers = n_layers

        assert self.mode in self.modes, "%s not in (%s, %s, %s, %s)" % (self.mode, *self.modes)
        self._init_input_representations()

        self.convs = torch.nn.ModuleList([torch.nn.Conv1d(self.embeddings_dim, self.filters, k) for k in self.kernel_sizes])
        self.projection = torch.nn.Linear(in_features=self.l*len(self.kernel_sizes)*self.filters, out_features=self.n_classes)
        self.dropout_layer = torch.nn.Dropout(self.dropout)
        self.build()

    def _init_input_representations(self):
        if self.mode == "trainable":
            self.embedding, self.tokenizer = get(model=self.representation, freeze=False)
            self.embeddings_dim = self.embedding.weight.shape[-1]
        elif self.mode == "untrainable":
            self.embedding, self.tokenizer = get(model=self.representation, freeze=True)
            self.embeddings_dim= self.embedding.weight.shape[-1]
        elif self.mode =="multichannel":
            self.l = 2
            self.embedding, self.tokenizer = get(model=self.representation, freeze=True)
            self.embeddings_dim = self.embedding.weight.shape[-1]

            self.embedding_untrainable = torch.nn.Embedding(*self.embedding.weight.shape)
            self.embedding_untrainable = self.embedding.from_pretrained(self.embedding.weight.clone(), freeze=False)
            self.embedding = torch.nn.ModuleList([self.embedding, self.embedding_untrainable])

        elif self.mode == "transformer":
            try:
                if self.n_layers == 1:
                    self.embedding, self.tokenizer = get(model=self.representation)
                    self.embeddings_dim = self.embedding(torch.tensor([[0]]))[0].shape[-1]
                else:
                    self.embedding, self.tokenizer = get(model=self.representation, output_hidden_states=True)
                    self.embeddings_dim = \
                    torch.cat(self.embedding(self.embedding.dummy_inputs["input_ids"])[2][self.n_layers:],
                              -1).shape[-1]
            except TypeError:
                print("If your using a model that does not support returning hiddenstates, set n_layers=1")
                import sys
                sys.exit()

    def forward(self, x):

        if self.mode == "trainable":
            embedded = self.embedding(x).permute(0, 2, 1)
            embedded = self.dropout_layer(embedded)
            c = [torch.nn.functional.relu(conv(embedded).permute(0, 2, 1).max(1)[0]) for conv in self.convs]
        elif self.mode == "untrainable":
            with torch.no_grad():
                embedded = self.embedding(x).permute(0, 2, 1)
            embedded = self.dropout_layer(embedded)
            c = [torch.nn.functional.relu(conv(embedded).permute(0, 2, 1).max(1)[0]) for conv in self.convs]
        elif self.mode =="multichannel":
            embedded_1 = self.embedding[0](x).permute(0, 2, 1)
            with torch.no_grad():
                embedded_2 = self.embedding[1](x).permute(0, 2, 1)
            embedded_1 = self.dropout_layer(embedded_1)
            embedded_2 = self.dropout_layer(embedded_2)
            c = [torch.nn.functional.relu(conv(embedded_1).permute(0, 2, 1).max(1)[0]) for conv in self.convs]+\
                [torch.nn.functional.relu(conv(embedded_2).permute(0, 2, 1).max(1)[0]) for conv in self.convs]
        elif self.mode == "transformer":

            if self.n_layers == 1:
                with torch.no_grad():
                    embedded = self.embedding(x)[0].permute(0, 2, 1)
            else:
                with torch.no_grad():
                    embedded = torch.cat(self.embedding(x)[2][(self.n_layers):], -1).permute(0, 2, 1)
            # embedded = self.dropout_layer(embedded)
            c = [torch.nn.functional.relu(conv(embedded).permute(0, 2, 1).max(1)[0]) for conv in self.convs]

        c = torch.cat(c, 1)
        output = self.projection(self.dropout_layer(c))
        return output



import numpy as np
from torch.nn.functional import one_hot
import torch

def makesequencelabels(query, maxlen, tagset):
    tagsetmap = dict(zip(tagset, range(len(tagset))))
    labels = [[tagsetmap[token] for token in lseq] for lseq in query]
    length = [len(x) for x in query]
    labels = [one_hot(x, len(tagset)) for x in labels]

    result = np.full((len(query), maxlen, len(tagset)), 0., dtype="float32")
    for i, e in enumerate(labels):
        result[i, :length[i], :] = e
    return np.stack(result)

def makemultilabels(query, maxlen, tagset=None):
    if tagset is not None:
        tagsetmap = dict(zip(tagset, range(len(tagset))))
        labels = [[tagsetmap[token] for token in lseq] for lseq in query]
    else:
        labels = query
    return torch.stack([torch.sum(one_hot(torch.LongTensor(x), maxlen), 0) for x in labels], dim=0)

def to_scheme(tagset, scheme="iobes", outside="O"):
    scheme = scheme.lower()
    if scheme == "iobes" or scheme == "bioes":
        return ([outside] + ["I-" + x for x in tagset] + ["B-" + x for x in tagset] + ["E-" + x for x in tagset] + [
            "S-" + x for x in tagset])
    elif (scheme == "bilou"):
        return (
                [outside] + ["B-" + x for x in tagset] + ["I-" + x for x in tagset] + ["L-" + x for x in tagset] + [
            "U-" + x for x in tagset])
    elif (scheme == "bio" or scheme == "iob"):
        return ([outside] + ["B-" + x for x in tagset] + ["I-" + x for x in tagset])
    elif (scheme == "noprefix"):
        return [outside] + tagset

def schemetransformer(column, scheme="BIOES", multilabel=False):
    scheme = scheme.upper()
    scheme = "BIOES" if scheme == "IOBES" else scheme
    data = column#self.data_manager.get_data(purpose=purpose)
    transforming_column = [tagset if isinstance(tagset, list) else [tagset] for tagset in data]

    newtags_column = []
    for i, tagset in enumerate(transforming_column):
        if "O" in tagset or "O" == tagset:
            newtags_column.append(tagset)
        elif scheme == "NOPREFIX":
            newtags_column.append([t.split("-")[-1] for t in tagset])
        else:
            currentTagset = [t.split("-")[-1] for t in tagset]
            newtags = []
            if i == 0:
                nextTagset = [t.split("-")[-1] for t in transforming_column[i + 1]]
                for t in currentTagset:
                    if t in nextTagset:
                        newtags_column.append("B-" + t)
                    else:
                        if scheme == "BIOES": newtags.append("S-" + t)
                        if scheme == "BIO1": newtags.append("I-" + t)
                        if scheme == "BIO2": newtags.append("B-" + t)
                        newtags_column.append(newtags)
            elif i == len(transforming_column) - 1:
                previousTagset = [t.split("-")[-1] for t in
                                  transforming_column[i - 1]]
                for t in currentTagset:
                    if t in previousTagset:
                        if scheme == "BIOES": newtags.append("E-" + t)
                        if scheme == "BIO1": newtags.append("I-" + t)
                        if scheme == "BIO2": newtags.append("I-" + t)
                    else:
                        if scheme == "BIOES": newtags.append("S-" + t)
                        if scheme == "BIO1": newtags.append("I-" + t)
                        if scheme == "BIO2": newtags.append("B-" + t)
                newtags_column.append(newtags)
            else:

                nextTagset = [t.split("-")[-1] for t in
                              transforming_column[i + 1]]
                previousTagset = [t.split("-")[-1] for t in
                                  transforming_column[i - 1]]
                newtags = []
                for t in currentTagset:
                    if t in previousTagset and t in nextTagset:
                        newtags.append("I-" + t)
                    elif t in previousTagset and t not in nextTagset:
                        if scheme == "BIOES": newtags.append("E-" + t)
                        if scheme == "BIO1": newtags.append("I-" + t)
                        if scheme == "BIO2": newtags.append("I-" + t)
                    elif t not in previousTagset and t in nextTagset:
                        newtags.append("B-" + t)
                    else:
                        if scheme == "BIOES": newtags.append("S-" + t)
                        if scheme == "BIO1": newtags.append("I-" + t)
                        if scheme == "BIO2": newtags.append("B-" + t)
                newtags_column.append(newtags)

    if not multilabel:
        newtags_column = [x[0] for x in newtags_column]
    return newtags_column


"""Layers that are concerned with scoring labels"""

import torch


class LabelAttention(torch.nn.Module):
    """Reimplementation of label attention as described in paper: Label-specific Document representations.
    Might be deprecated.
    """
    def __init__(self, n_classes, input_dim, hidden_dim, label_repr=None, freeze=True):
        super(LabelAttention, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.n_classes = n_classes

        if self.hidden_dim is not None:
            if self.hidden_dim != self.input_dim:
                self.projection = torch.nn.Linear(self.input_dim, self.hidden_dim)

        if label_repr is None:
            self.label_repr = torch.nn.Parameter(torch.Tensor(n_classes, self.hidden_dim))
            torch.nn.init.kaiming_normal_(self.label_repr)
        else:
            assert label_repr.shape[-1] == hidden_dim,"label embedding dimension must equal hidden_dim"
            self.label_repr = torch.nn.Parameter(label_repr)
            self.label_repr.requires_grad=freeze

    def forward(self, x):
        if self.hidden_dim is not None:
            if self.hidden_dim != self.input_dim:
                x = self.projection(x)
        A =torch.softmax(torch.matmul(x, self.label_repr.permute(1,0)),-1)
        output = torch.matmul(A.permute(0,2,1), x)
        return output, A

class LabelEmbeddingAttention(torch.nn.Module):
    def __init__(self, n_classes, input_dim, hidden_dim, label_embedding):
        super(LabelEmbeddingAttention, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.n_classes = n_classes

        if self.hidden_dim is not None:
            if self.hidden_dim != self.input_dim:
                self.projection = torch.nn.Linear(self.input_dim, self.hidden_dim)
        self.label_repr = label_embedding

    def forward(self, x):
        A =  torch.softmax(torch.matmul(x, self.label_repr.permute(1,0)),-1)
        output = torch.matmul(A.permute(0,2,1), x)
        return output, A


class LabelEmbeddingScoring(torch.nn.Module):
    """Layer that keeps a representation (static Embedding) and compares the input to all vectors. The metric
        should be freely choosable
    """
    def __init__(self, n_classes, input_dim, label_repr, similarity="cosine", label_freeze=True):
        super(LabelEmbeddingScoring, self).__init__()
        self.input_dim = input_dim
        self.n_classes = n_classes

        assert similarity in ["cosine","euclidean"], "Distance metric %s not implemented." % (similarity, )
        self.similarity=similarity

        self.label_repr = torch.nn.Parameter(torch.from_numpy(label_repr).float())
        self.label_repr.requires_grad=not label_freeze
        self.projection = torch.nn.Linear(self.input_dim, self.label_repr.shape[-1])

    def forward(self, x):
        x = self.projection(x)
        if self.similarity=="cosine":
            output = torch.matmul(
                x/torch.norm(x,p=2,dim=-1).unsqueeze(-1),
                (self.label_repr/torch.norm(self.label_repr, p=2,dim=-1).unsqueeze(-1)).transpose(0,1)
            )
        if self.similarity=="euclidean":
            output = torch.sigmoid(
                torch.norm((x.unsqueeze(2) - self.label_repr.unsqueeze(0).unsqueeze(1)),p=2,dim=-1)
            )
        return output

class LabelSpecificSelfAttention(torch.nn.Module):
    def __init__(self, n_classes, input_dim, hidden_dim):
        super(LabelSpecificSelfAttention, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.n_classes = n_classes

        # self.to_hidden = torch.nn.Parameter(torch.Tensor(self.input_dim, self.hidden_dim))
        self.to_label = torch.nn.Parameter(torch.Tensor(self.hidden_dim, self.n_classes))

        self.to_hidden = torch.nn.Linear(self.input_dim, self.hidden_dim)
        # self.to_label = torch.nn.Linear(self.hidden_dim, self.n_classes)

        # torch.nn.init.kaiming_normal_(self.to_hidden)
        torch.nn.init.kaiming_normal_(self.to_label)

    def forward(self,x):
        att = torch.softmax(torch.matmul(
            torch.tanh(self.to_hidden(x)),self.to_label),
            -1)
        # att = torch.softmax(self.to_label(torch.tanh(self.to_hidden(x))), -1)
        return torch.matmul(att.permute(0,2,1),x), att

class AdaptiveCombination(torch.nn.Module):
    def __init__(self, input_dim, n_classes ):
        super(AdaptiveCombination, self).__init__()
        self.input_dim = input_dim
        self.n_classes = n_classes

        self.alpha_weights = torch.nn.Linear(input_dim,1)
        self.beta_weights = torch.nn.Linear(input_dim,1)
        # torch.nn.init.kaiming_normal_(self.alpha_weights)
        # torch.nn.init.kaiming_normal_(self.beta_weights)

    def forward(self, x):
        # alpha = torch.sigmoid(torch.matmul(x[0], self.alpha_weights))
        # beta = torch.sigmoid(torch.matmul(x[1], self.beta_weights))

        alpha = torch.sigmoid(self.alpha_weights(x[0]))
        beta = torch.sigmoid(self.beta_weights(x[1]))

        #constrain the sum to one
        alpha = alpha / (alpha + beta)
        beta = beta / (alpha + beta)
        output = alpha*x[0] + beta*x[1]
        return output



import gzip
import tempfile
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile

import networkx as nx
import rdflib
from rdflib import Graph as RDFGraph
from rdflib.extras.external_graph_libs import rdflib_to_networkx_graph
from tqdm import tqdm

from ..data.data_loaders import _save_to_tmp, _load_from_tmp


def transform(x, rg, lang="en"):
    if isinstance(x, rdflib.term.URIRef):
        from rdflib.namespace import SKOS
        if str(SKOS) in str(x): return [x.split("#")[-1]]
        else:
            return [str(l[1]) for l in rg.preferredLabel(x) if l[1].language==lang]
    elif isinstance(x, rdflib.term.Literal):
        if x.language != lang: return None
        else: return [str(x)]
    elif isinstance(x, rdflib.term.BNode):
        return None#[str(x)]
    return (x)


def transform_triples(rg,lang="en"):
    new_list =  [[transform(x, rg, lang) for x in t] for t in tqdm(rg)]
    new_list = [x for x in new_list
                if (not (any([r==[] for r in x]) or any([r is None for r in x]))) and ( "prefLabel" not in x[1])]
    return new_list



def load_mesh():

    url = "ftp://ftp.nlm.nih.gov/online/mesh/rdf/2020/"
    resp = urlopen(url+"mesh2020.nt.gz")
    f = gzip.open(resp, mode="rb")
    content = f.read().decode()
    rg = RDFGraph()
    with tempfile.TemporaryDirectory() as tmpdir:
        print(tmpdir)
        with open(tmpdir + "/tst.nt", "w") as f: f.write(content)
        rg = rg.parse(tmpdir + "/tst.nt", format='nt')

    G = rdflib_to_networkx_graph(rg)
    print("rdflib Graph loaded successfully with {} triples".format(len(rg)))
    return G

def load_nasa():
    data = _load_from_tmp("nasa")
    if data is not None:
        return data
    else:
        url = "https://www.sti.nasa.gov/download/NASA_Thesaurus_Alpha_SKOS.xml"
        response = urlopen(url)
        content = response.read().decode("utf-16")
        rg = RDFGraph()
        rg = rg.parse(data=content)
        literal_triples = transform_triples(rg)
        g = nx.DiGraph()
        for triple in literal_triples:
            for u in triple[0]:
                for v in triple[2]:
                    for p in triple[1]:
                        g.add_edge(u, v, label=p)
        _save_to_tmp("nasa", g)
        return g



def load_gesis():
    data = _load_from_tmp("gesis")
    if data is not None:
        return data
    else:
        url = "http://lod.gesis.org/thesoz-komplett.xml.gz"
        response = urlopen(url)
        f = gzip.open(response, mode="rb")
        content = f.read().decode()
        rg = RDFGraph()
        rg = rg.parse(data = content)
        literal_triples = transform_triples(rg)
        g = nx.DiGraph()
        for triple in literal_triples:
            for u in triple[0]:
                for v in triple[2]:
                    for p in triple[1]:
                        g.add_edge(u, v, label = p)
        _save_to_tmp("gesis",g)
        return g

def load_stw():
    data = _load_from_tmp("stw")
    if data is not None:
        return data
    else:
        url = "https://aspra29.informatik.uni-leipzig.de:9090/stw.rdf.zip"
        response = urlopen(url)
        zf = ZipFile(BytesIO(response.read()))
        with zf.open("stw.rdf") as f:
            content = f.read()
        rg = RDFGraph()
        rg = rg.parse(data=content)
        literal_triples = transform_triples(rg)
        g = nx.DiGraph()
        for triple in literal_triples:
            for u in triple[0]:
                for v in triple[2]:
                    for p in triple[1]:
                        g.add_edge(u, v, label=p)
        _save_to_tmp("stw",g)
        return g

def load_elsevier():
    """https://www.elsevier.com/__data/assets/pdf_file/0010/175249/ACAD_FP_FS_FPEFactSheet2019_WEB.pdf"""
    print("Be Careful. This is not yet the full elsevier thesaurus. ONly gesis, nasa and stw")
    gesis = load_gesis()
    nasa = load_nasa()
    stw = load_stw()
    c = nx.compose(nx.compose(gesis, nasa), stw)
    return c

def load_wordnet():
    """
    Loading the wordnet graph as a networkx.DiGraph
    :return: A networkx.DiGraph containing wordnet.
    """
    try:
        from nltk.corpus import wordnet as wn
    except:
        print("To use this function you have to install nltk.")
    G = nx.DiGraph()
    for ss in wn.all_synsets():
        if ss.hypernyms() != []:
            for hypernym in ss.hypernyms():
                for hh in hypernym.lemmas():
                    for word in ss.lemmas():
                        G.add_edge(hh.name().replace("_"," "), word.name().replace("_"," "), label="related")

    return G


def load_wordnet_sample(num=1000):
    """
    Loading the wordnet graph as a networkx.DiGraph
    :num: The size of the subsample ( Only the first n synsets will be added )
    :return: A networkx.DiGraph containing wordnet.
    """
    try:
        from nltk.corpus import wordnet as wn
    except:
        print("To use this function you have to install nltk.")
    G = nx.DiGraph()
    i=0
    for ss in wn.all_synsets():
        if ss.hypernyms() != []:
            for hypernym in ss.hypernyms():
                for hh in hypernym.lemmas():
                    for word in ss.lemmas():
                        G.add_edge(hh.name().replace("_"," "), word.name().replace("_"," "), label="related")
                        i += 1
        if i >=num:
            break
    return G


def load_NELL():
    """
    Download and return the NELL Graph (Never Ending Lanuage Learning
    Will return only elements that have a confidence of 0.999 and larger.
    :return: A networkx.DiGraph
    """
    url = "http://rtw.ml.cmu.edu/resources/results/08m/NELL.08m.1115.esv.csv.gz"
    data = _load_from_tmp("NELL")
    if data is not None:
        return data
    else:
        with tempfile.TemporaryDirectory() as tmpdir:
            resp = urlopen(url)
            tf = gzip.open(resp, mode="r")
            content = tf.read()
            x = content.decode().split("\n")

            import networkx as nx
            G = nx.DiGraph()
            for line in x[1:]:
                test = line.split("\t")
                if len(test) < 10: continue
                confidence = test[4]
                if float(confidence) > 0.999:
                    relation = test[1]
                    if relation != "generalizations" and relation != "concept:latitudelongitude":
                        entityLiterals = test[8].lower()
                        valueLiterals = test[9].lower()
                        G.add_edge(entityLiterals.replace("_"," "), valueLiterals.replace("_"," "), label=relation, confidence=confidence)
                    if relation == "generalizations":
                        entityLiterals = test[8].lower()
                        valueLiterals = test[2].lower().split(":")[-1]
                        G.add_edge(entityLiterals.replace("_"," "), valueLiterals.replace("_"," "), label=relation, confidence=confidence)
        _save_to_tmp("NELL",G)
        return G



from torch.utils.data import Dataset

class RawTextDatasetTokenizer(Dataset):
    """
    Dataset to hold raw text and tokenize batches.
    """
    def __init__(self, x, tokenizer, length=512, bidirectional=False, **kwargs):
        self.__dict__.update(kwargs)
        self.x = x
        self.tokenizer = tokenizer
        self.bidirectional=bidirectional
        self.tokenized=self.tokenizer.encode(x).tokens
        self.length = length

    def __len__(self):
        return len(self.tokenized) - self.length - 2

    def __getitem__(self, idx):
        r = {'input': self.tokenized[(idx+1):(idx+self.length+1)],
             'forward': self.tokenized[(idx+1+1):(idx+self.length+1+1)]}
        if self.bidirectional:
            r['backward'] = self.tokenized[(idx):(idx+self.length)]
        return r

class RawTextDataset(Dataset):
    """
    Dataset to hold raw text and tokenize batches.
    """
    def __init__(self, x, target="words", length=512, bidirectional=False, **kwargs):
        self.__dict__.update(kwargs)
        if callable(target):
            self.x = target(x)
        elif target=="words":
            self.x = x.split()
        elif target =="character":
            self.x = list(x)
        else:
            print("target has to be either 'words' or 'character' or a callable function that splits the input string.")

        self.bidirectional=bidirectional
        self.length = length

    def __len__(self):
        return len(self.x) - self.length - 2

    def __getitem__(self, idx, next=True):
        r = {'input': "".join(self.x[(idx+1):(idx+self.length+1)])}
        if next:
            r['forward'] = self.x[(idx+self.length+1)]
        else:
            r['forward'] = self.x[(idx + 1 + 1):(idx + self.length + 1 + 1)]
            if self.bidirectional:
                r['backward'] = self.x[(idx):(idx+self.length)]
        return r

    def generate_wordlist(self, n=None):
        from collections import Counter
        c = Counter(self.x)
        if n is None:
            n=len(c)
        return [x[0] for x in c.most_common(n)]

from torch import tensor
class RawTextDatasetTensor(Dataset):
    """
    Dataset to hold raw text and tokenize batches.
    """
    def __init__(self, x, vocab, target="words", length=512, bidirectional=False, **kwargs):
        self.__dict__.update(kwargs)
        if callable(target):
            self.x = target(x)
        elif target=="words":
            self.x = x.split()
        elif target =="character":
            self.x = list(x)
        else:
            print("target has to be either 'words' or 'character' or a callable function that splits the input string.")
        self.window = length+1
        self.vocab = dict(zip(vocab, range(len(vocab))))

        sequences = []
        for start in range(0, len(self.x) - self.window):
            sequences.append([x for x in self.x[start:start + self.window]])
        self.x = sequences

        self.tr = tensor([[self.vocab.get(c, len(vocab)+1) for c in s] for s in self.x]).long()

    def __len__(self):
        return self.tr.shape[0]

    def __getitem__(self, idx, raw=False):
        resource = self.x if raw else self.tr
        return resource[idx,:-1], resource[idx,-1]


# https://github.com/shenweichen/GraphEmbedding

import networkx as nx
import numpy as np
from node2vec import Node2Vec
from sklearn import random_projection
from sklearn.decomposition import NMF


def subgraph_extract(X, graph, subnodelist):
    new = np.zeros_like(X)
    for i, nm in enumerate(graph.nodes):
        if nm in subnodelist.keys():
            new[subnodelist[nm]] = X[i]
    return new


def get_node2vec(graph, classes, dim, return_all=False):
    """
    Embed a graph using the node2vec algorithm
    :param graph: A networkx graph
    :param classes: Classes to be extracted from the graph. (The subset of nodes for which the embedding will be returned)
    :param dim: The dimension of the embedding
    :param return_all: If True the embedding to every node will be returned. If False only the subset of embeddings for nodes that are in 'classes' will be returned.
    :return: A Matrix of embedding vectors
    """
    node2vec = Node2Vec(graph, dimensions=dim, walk_length=30, num_walks=500, workers=4)
    model = node2vec.fit(window=50, min_count=1, batch_words=4)
    W = model.wv.vectors
    if return_all: W=subgraph_extract(W,graph, dict(zip(list(graph.nodes), range(len(graph.nodes)))))
    else: W=subgraph_extract(W,graph,classes)
    return W

def get_nmf(graph, classes, dim, return_all=False):
    """
    Embed a graph using the non-negative matrix factorization algorithm.
    :param graph: A networkx graph
    :param classes: Classes to be extracted from the graph. (The subset of nodes for which the embedding will be returned)
    :param dim: The dimension of the embedding
    :param return_all: If True the embedding to every node will be returned. If False only the subset of embeddings for nodes that are in 'classes' will be returned.
    :return: A Matrix of embedding vectors
    """
    model = NMF(n_components=dim, init='random', random_state=0)
    W = model.fit_transform(nx.to_numpy_array(graph))
    H = model.components_
    if return_all: W=subgraph_extract(W,graph, dict(zip(list(graph.nodes), range(len(graph.nodes)))))
    else: W=subgraph_extract(W,graph,classes)
    return W

def get_random_projection(graph, classes, dim, return_all=False):
    """
    Embed a graph using the random projection algorithm.
    :param graph: A networkx graph
    :param classes: Classes to be extracted from the graph. (The subset of nodes for which the embedding will be returned)
    :param dim: The dimension of the embedding
    :param return_all: If True the embedding to every node will be returned. If False only the subset of embeddings for nodes that are in 'classes' will be returned.
    :return: A Matrix of embedding vectors
    """
    transformer = random_projection.GaussianRandomProjection(n_components=dim)
    W = transformer.fit_transform(nx.to_numpy_array(graph))
    if return_all: W=subgraph_extract(W,graph, dict(zip(list(graph.nodes), range(len(graph.nodes)))))
    else: W=subgraph_extract(W,graph,classes)
    return W



"""
Functions and classes for loading and using numerical text representations like word embeddings
and language models in an automated fashion.

Most of the time this module will only be used implicitly when creating an instance of an classification model.

## mlmc.representation.get

The main function of `mlmc.representation` is `mlmc.representation.get`. It is a wrapper around downloading,formatting
 and initializting word sequence embedding techniques.
 It works as a string lookup and should support  all model names found on
 https://huggingface.co/models (currently only tested for  common non-community models)
 or one of the keywords ["glove50", "glove100", "glove200" and "glove300"] for static
 embeddings.

 The function always returns an embedding object that inherits from `torch.nn.Module` (so it can be used in
 your own model architectures as layers) and the corresponding tokenizer, so that calling both sequentially will result
in the embedded version of your string, but gives you the freedom to use your own tokenizer ( although not recommended in
case you are using the language models).

Getting an embedding example

```python
import mlmc
embedder, tokenizer = mlmc.representation.get("bert-base-uncased")
embedding_tensor = embedder(tokenizer("Some text you want to embed"))
```
"""

from .representations import load_static, get_transformer, map_vocab,get_embedding,get,get_white_space_tokenizer, is_transformer, empty_cache, custom_embedding, delete_custom_embeddings
from .character import charindex
from .labels import makesequencelabels, schemetransformer, to_scheme, makemultilabels
from .postprocessing_vectors import postprocess_embedding
from .label_embeddings import get_word_embedding_mean, get_lm_generated, get_lm_repeated, get_graph_augmented
from .embedder import Embedder

from .output_transformations import threshold_mcut, threshold_hard, threshold_max


"""
Few-Shot and Zero-Shot Multi-Label Learning for Structured Label Spaces - Rios & Kavuluru (2018)
"""
import torch
from ..models.abstracts import TextClassificationAbstract
from ..representation import get, is_transformer
import re

class ZAGCNNLM(TextClassificationAbstract):
    def __init__(self, classes,   adjacency, method, scale, representation="roberta", max_len=200, dropout = 0.5, norm=False, n_layers=4, **kwargs):
        super(ZAGCNNLM, self).__init__(**kwargs)

        self.classes = classes
        self.n_classes = len(classes)
        self.max_len = max_len
        self.use_dropout = dropout
        self.filters = 300
        self.hidden_dim=512
        self.scale = scale
        self.kernel_sizes = [3,10]
        self.dropout = dropout
        self.adjacency = adjacency
        self.method = method
        self.adjacency_param = torch.nn.Parameter(torch.from_numpy(adjacency).float())
        self.adjacency_param.requires_grad = False
        self.n_layers=n_layers
        self.norm = norm
        self.representation = representation
        self._init_input_representations()

        from ..graph import get as gget
        self.graph = gget(["stw"])
        self.create_labels(classes, method=self.method, scale=self.scale)

        self.convs = torch.nn.ModuleList(
            [torch.nn.Conv1d(self.embeddings_dim, self.filters, k) for k in self.kernel_sizes])
        self.pool = torch.nn.MaxPool1d(3, stride=2)
        self.document_projection = torch.nn.Linear(self.filters, self.label_embeddings_dim)


        self.dropout_layer= torch.nn.Dropout(self.dropout)
        import torch_geometric as torchg
        self.gcn1 = torchg.nn.GCNConv(in_channels=self.label_embeddings.shape[-1], out_channels=self.hidden_dim)
        self.gcn2 = torchg.nn.GCNConv(in_channels=self.hidden_dim, out_channels=self.hidden_dim)
        self.projection = torch.nn.Linear(in_features=self.filters, out_features=self.hidden_dim+self.label_embeddings.shape[-1])
        self.build()

    def forward(self, x):
        if self.n_layers == 1:
            with torch.no_grad():
                embeddings = self.embedding(x)[0]
        else:
            with torch.no_grad():
                embeddings = torch.cat(self.embedding(x)[2][-self.n_layers:], -1)
        embedded = self.dropout_layer(embeddings)
        c = torch.cat([self.pool(torch.nn.functional.relu(conv(embedded.permute(0,2,1)))) for conv in self.convs], dim=-1).permute(0,2,1)
        d2 = torch.tanh(self.document_projection(c))
        if self.norm: d2 = d2/d2.norm(p=2,dim=-1,keepdim=True)
        a = torch.softmax(torch.matmul(d2, self.label_embeddings.t()), -1)
        label_wise_representation = torch.matmul(a.permute(0, 2, 1), c)

        label_wise_representation = self.dropout_layer(label_wise_representation)

        labelgcn = self.gcn1(self.label_embeddings, torch.stack(torch.where(self.adjacency_param == 1), dim=0))
        labelgcn = self.dropout_layer(labelgcn)
        labelgcn = self.gcn2(labelgcn, torch.stack(torch.where(self.adjacency_param == 1), dim=0))
        labelvectors = torch.cat([self.label_embeddings, labelgcn], dim=-1)
        return (torch.relu(self.projection(label_wise_representation)) * labelvectors).sum(-1)

    def create_labels(self, classes, method="repeat",scale="mean"):
        # assert method in ("repeat","generate","embed", "glove", "graph"), 'method has to be one of ("repeat","generate","embed")'
        self.classes = classes
        if method=="repeat":
            from ..representation import get_lm_repeated
            l = get_lm_repeated(self.classes, self.representation)
        if method == "generate":
            from ..representation import get_lm_generated
            l = get_lm_generated(self.classes, self.representation)
        if method == "embed":
            l = self.embedding(self.tokenizer(self.classes.keys()).to(list(self.parameters())[0].device))[1]
        if method == "glove":
            from ..representation import get_word_embedding_mean
            l = get_word_embedding_mean(
                [" ".join(re.split("[/ _-]", x.lower())) for x in self.classes.keys()],
                "glove300")
        if method == "graph":
            from ..representation import get_graph_augmented
            l = get_graph_augmented(self.classes, graph=self.graph, model=self.representation, topk=200, batch_size=64, device=self.device)

        if scale=="mean":
            print("subtracting mean")
            l = l - l.mean(0,keepdim=True)
        if scale == "normalize":
            print("normalizing")
            l = l/l.norm(p=2,dim=-1,keepdim=True)

        self.label_embeddings = torch.nn.Parameter(l.to(self.device))
        self.label_embeddings.requires_grad = False
        self.label_embeddings_dim = self.label_embeddings.shape[-1]

    def set_adjacency(self, adj):
        del self.adjacency_param
        self.adjacency_param = torch.nn.Parameter(torch.from_numpy(adj).float().to(self.device))
        self.adjacency_param.requires_grad = False


"""
A collection of sampling methods for the MultilabelDataset object

.
"""
import numpy as np

def sampler(dataset, fraction=None, absolute=None):
    """
    Sample a Random subsample of fixed size or fixed fraction of a dataset (i.i.d. sample).
    Args:
        dataset: A instance of mlmc.data.MultilabelDataset
        fraction: The fraction of the data that should be returned (0<fraction<1)
        absolute: The absolute size of the sampled dataset
    Returns:
         A randomly subsampled MultilabelDataset of the desired size.
    """
    
    # assert fraction is None != absolute is None, "Exactly one of fraction or absolute has to be set."
    if fraction is not None:
        assert fraction<1 and fraction>0, "The fraction argument has to be between 0 and 1."
    n_samples = absolute if absolute is not None else  int(fraction*len(dataset))
    if n_samples > len(dataset): n_samples = len(dataset)
    ind = np.random.choice(range(len(dataset)), n_samples, replace=False)
    x = [dataset.x[i] for i in ind]
    y = [dataset.y[i] for i in ind]
    return type(dataset)(x=x, y=y, classes=dataset.classes, target_dtype=dataset.target_dtype)


def successive_sampler(dataset, classes, separate_dataset, reindex_classes=True):
    """
    Return an iterable of datasets sampled from dataset.

    Args:
         dataset: The input dataset (MultilabelDataset)
         classes: A classes mapping
         separate_dataset: The number of datasets to generate
         reindex_classes: If True, classes in the subsampled datasets are reindexed to 1:len(classes)
    Returns:
        A list of MultilabelDatasets
    """

    # Quick FIX function was changing the global value of mutable
    # ToDO: Find a better way maybe than copying
    classes=classes.copy()

    n_result = []
    n_idx = []
    n_ind = []
    already_select_id = set()
    already_select_class = {}
    n_samples =  np.round(len(dataset)/separate_dataset).astype(np.int64)

    for x in range(0, separate_dataset):
        # extend classes to use
        # sample from existing classes/delete from whole dataset to sample
        c_ind = np.random.choice(range(len(classes)), np.round(len(classes)/2).astype(np.int64))
        selectedKeys = list()
        # Iterate over the dict and put to be deleted keys in the list
        for index, (key, value) in enumerate(classes.items()):
            if(index in c_ind):
                print(index, key, value)
                already_select_class[key] = value
                selectedKeys.append(key)
 
        #Iterate over the list and delete corresponding key from dictionary
        for key in selectedKeys:
            if key in classes :
                del classes[key]
        l_list = dataset.y
        
        #Object to store which data point belongs to already defined classes
        candidate_idx = []

        for index, (c_list) in enumerate(dataset.y):
            intersect = set(c_list).intersection(set(already_select_class.keys()))
            if len(intersect) > 0:
                candidate_idx.append(index)
                l_list[index] = list(intersect)

        ind = np.random.choice(list(set(candidate_idx) - already_select_id), n_samples)

        already_select_id = set(ind).union(already_select_id)
        x = [dataset.x[i] for i in list(already_select_id)]
        y = [l_list[i] for i in list(already_select_id)]

        x_new = [dataset.x[i] for i in ind]
        y_new = [l_list[i] for i in ind]

        # Quick FIX function was changing the global value of mutable
        # ToDO: Find a better way maybe than copying
        cls = dict(zip(already_select_class.keys(), range(len(already_select_class)))) if reindex_classes else already_select_class.copy()

        n_result.append({
            'train' : type(dataset)(x=x, y=y, classes=dataset.classes, occuring_classes=cls, target_dtype=dataset.target_dtype),
            'test':type(dataset)(x=x_new, y=y_new, classes=dataset.classes, occuring_classes=cls,target_dtype=dataset.target_dtype)
        })
        n_idx.append(list(already_select_id))

        print("--- LENGTH DATASET: ",len(already_select_id))

    return n_result,n_idx


def class_sampler(data, classes, samples_size):
    """
    Create a subset of the data containing only classes in the classes argument and subsampling each class
    roughly to the number given in sample_size.

    Args:
        data: MultilabelDataset to sample from
        classes: subset of classes
        samples_size: Number of examples for each class (roughly)
    Returns:
        A dataset with roughly sample_size number of instances per class in classes.
    """
    if isinstance(samples_size, int):
        n = [samples_size for _ in range(len(classes))]
    else:
        n = samples_size

    index_list = list(range(len(data.x)))
    import random
    random.shuffle(index_list)

    new_data = []
    new_labels = []
    size_counter = [0 for _ in range(len(classes))]

    for i in index_list:
        example = data.x[i]
        labels = data.y[i]
        if any([x in classes for x in labels]):
            current_labels = [x for x in labels if x in classes]

            fraction_of_needed_samples = sum([size_counter[classes.index(x)] for x in current_labels]) /\
                                         (1+sum([min(n[classes.index(x)], size_counter[classes.index(x)] ) for x in current_labels]))

            if fraction_of_needed_samples <=1:
                new_data.append(example)
                new_labels.append(current_labels)

                for c in current_labels:
                    size_counter[classes.index(c)] = 1 + size_counter[classes.index(c)]

        if all([True if i>=j else False for i,j in zip(size_counter,n)]):
            break

    return type(data)(x=new_data, y=new_labels, classes=data.classes, occurring_classes={k: data.classes[k] for k in classes})



def validation_split(dataset, fraction=None, absolute=None):
    """
    Split a dataset into two separate splits (validation split)

    Args:
        dataset: data set to split
        fraction: The fraction of the data that should be returned (0<fraction<1)
        absolute: The absolute size of the sampled dataset
    Returns:
         A tuple of randomly subsampled MultilabelDatasets of the desired size.
    """
    # assert fraction is None != absolute is None, "Exactly one of fraction or absolute has to be set."
    if fraction is not None:
        assert fraction < 1 and fraction > 0, "The fraction argument has to be between 0 and 1."
    ind = list(range(len(dataset)))
    np.random.shuffle(ind)
    n_samples = (len(dataset)-absolute) if absolute is not None else int((1-fraction) * len(dataset))
    return type(dataset)(x=[dataset.x[i] for i in ind[:n_samples]],
                         y=[dataset.y[i] for i in ind[:n_samples]],
                         classes=dataset.classes), \
           type(dataset)(x=[dataset.x[i] for i in ind[n_samples:]],
                         y=[dataset.y[i] for i in ind[n_samples:]],
                         classes=dataset.classes)


import torch

class AttentionWeightedAggregation(torch.nn.Module):
    """
    Aggregate a tensor weighted by weights obtained from self attention
    https://arxiv.org/pdf/1703.03130.pdf
    """
    def __init__(self, in_features, d_a):
        super(AttentionWeightedAggregation, self).__init__()
        self.in_features=in_features
        self.d_a = d_a

        self.att_projection = torch.nn.Linear(in_features=self.in_features, out_features=self.d_a, bias=True)
        self.att_weights = torch.nn.Parameter(torch.zeros((self.d_a,)))
        torch.nn.init.normal_(self.att_weights, mean=0.0, std=0.001)


    def forward(self, x, aggr=None, return_att=True):
        attention = torch.softmax(torch.matmul(torch.tanh(self.att_projection(x)), self.att_weights), -1)
        if aggr is None:
            output = (attention[:, :, None]*x).sum(-2)
        else:
            output = (attention[:, :, None]*aggr).sum(-2)
        if return_att:
            return output, attention
        else:
            return output




import torch

def threshold_hard(x, tr=0.5):
    """
    Transform input tensor into a multiple-hot tensor using a threshold.

    Args:
        x: Tensor
        tr: Threshold

    Returns:
        Multiple-Hot Tensor
    """
    return (x > tr).int()

def threshold_mcut(x):
    """
    Transform input tensor into a multiple-hot tensor using a threshold

    The threshold is estimated from the largest decay in probability between two successively ranked labels.

    Args:
        x: Input Tensor

    Returns:
        Multiple-Hot Tensor
    """
    x_sorted = torch.sort(x, -1)[0]
    thresholds = (x_sorted[:, 1:] - x_sorted[:, :-1]).max(-1)[0]
    return (x > thresholds[:, None]).float()

def threshold_max(x):
    return torch.zeros_like(x).scatter(1, torch.topk(x, k=1)[1], 1)


import torch
import torch.nn as nn
from torch.nn.utils.rnn import PackedSequence
from typing import *


class VariationalDropout(nn.Module):
    """
    Applies the same dropout mask across the temporal dimension
    See https://arxiv.org/abs/1512.05287 for more details.
    Note that this is not applied to the recurrent activations in the LSTM like the above paper.
    Instead, it is applied to the inputs and outputs of the recurrent layer.
    """

    def __init__(self, dropout: float, batch_first: Optional[bool] = False):
        super().__init__()
        self.dropout = dropout
        self.batch_first = batch_first

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.dropout <= 0.:
            return x

        is_packed = isinstance(x, PackedSequence)
        if is_packed:
            x, batch_sizes, sorted_indices, unsorted_indices = x
            max_batch_size = int(batch_sizes[0])
        else:
            batch_sizes = None
            max_batch_size = x.size(0)

        # Drop same mask across entire sequence
        if self.batch_first:
            m = x.new_empty(max_batch_size, 1, x.size(2), requires_grad=False).bernoulli_(1 - self.dropout)
        else:
            m = x.new_empty(1, max_batch_size, x.size(2), requires_grad=False).bernoulli_(1 - self.dropout)
        x = x.masked_fill(m == 0, 0) / (1 - self.dropout)

        if is_packed:
            return PackedSequence(x, batch_sizes, sorted_indices, unsorted_indices)
        else:
            return x


class LSTM(nn.LSTM):
    """A copy of the better LSTM github repository"""
    def __init__(self, *args, dropouti: float = 0.,
                 dropoutw: float = 0., dropouto: float = 0.,
                 batch_first=True, unit_forget_bias=True, **kwargs):
        super().__init__(*args, **kwargs, batch_first=batch_first)
        self.unit_forget_bias = unit_forget_bias
        self.dropoutw = dropoutw
        # self.input_drop = torch.nn.Dropout(dropouti)
        # self.output_drop = torch.nn.Dropout(dropouto)
        self._init_weights()

    def _init_weights(self):
        """
        Use orthogonal init for recurrent layers, xavier uniform for input layers
        Bias is 0 except for forget gate
        """
        for name, param in self.named_parameters():
            if "weight_hh" in name:
                nn.init.orthogonal_(param.data)
            elif "weight_ih" in name:
                nn.init.xavier_uniform_(param.data)
            elif "bias" in name and self.unit_forget_bias:
                nn.init.zeros_(param.data)
                param.data[self.hidden_size:2 * self.hidden_size] = 1
        self.flatten_parameters()
    def _drop_weights(self):
        for name, param in self.named_parameters():
            if "weight_hh" in name:
                getattr(self, name).data = \
                    torch.nn.functional.dropout(param.data, p=self.dropoutw,
                                                training=self.training)
        self.flatten_parameters()

    def forward(self, input, hx=None):
        self._drop_weights()
        seq, state = super().forward(input, hx=hx)
        return seq, state




class LSTMRD(torch.nn.Module):
    """
    A self implementation of the LSTM with a tensorflow like recurrent dropout on the activations of the
    reccurrent cell
    """
    def __init__(self, input_size, hidden_size, bias=True,
                 batch_first=False, dropout=0.5, recurrent_dropout=0.5, bidirectional=False):
        super(LSTMRD,self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = torch.nn.Dropout(dropout)
        self.recurrent_dropout = torch.nn.Dropout(recurrent_dropout)
        self.bidirectional = bidirectional

        self.forward_cell = torch.nn.LSTMCell(self.input_size,self.hidden_size,bias=self.bias)

        if bidirectional:
            self.backward_cell  = torch.nn.LSTMCell(self.input_size,self.hidden_size,bias=self.bias)

    def forward(self, x, hidden=None):
        if self.batch_first:
            x = x.permute(1,0,2)
        if self.dropout.p>0:
            x = self.dropout(x)
        output = []
        hx = hidden[0][0]
        cx = hidden[1][0]
        for i in range(x.shape[0]):
            hx, cx = self.forward_cell(x[i], (hx, cx))
            if self.recurrent_dropout.p >0:
                cx = self.recurrent_dropout(cx)
            output.append(hx)
        forward = torch.stack(output,dim=0)
        if self.bidirectional:
            x = x.flip(0)
            hx = hidden[0][1]
            cx = hidden[1][1]
            output = []
            for i in range(x.shape[0]):
                hx, cx = self.backward_cell(x[i], (hx, cx))
                if self.recurrent_dropout.p>0:
                    cx = self.recurrent_dropout(cx)
                output.append(hx)
            output = torch.stack(output, dim=0)
            forward = torch.cat([forward, output.flip(0)],dim=-1)

        if self.batch_first:
            forward = forward.permute(1,0,2)
        return forward



import sklearn.metrics as skm
import torch


class MultiLabelReport():
    """Multilabel iterative F1/Precision/Recall. Ignite API like"""
    def __init__(self, classes, trf=lambda x : x,  check_zeros=False):
        self.classes = classes
        self.truth = []
        self.pred = []
        self.check_zeros = check_zeros
        self.trf = trf

    def update(self, batch):
        assert isinstance(batch, tuple), "batch needs to be a tuple"
        if self.check_zeros:
            non_zero_rows = (((batch[1]==0).sum(-1)==batch[1].shape[-1]).int()) ==0
            self.truth.append(batch[1][non_zero_rows])
            self.pred.append(batch[0][non_zero_rows])
        else:
            self.truth.append(batch[1])
            self.pred.append(batch[0])
    def compute(self):
        pred = self.trf(torch.cat(self.pred))
        return skm.classification_report((torch.cat(self.truth)).numpy(),
                                         pred.numpy(),
                                         output_dict=True,
                                         target_names=list(self.classes.keys()))

class AUC_ROC():
    """Multilabel iterative AUC_ROC. Ignite API like"""
    def __init__(self,n_classes, return_roc=False, reduction="macro"):
        self.values = torch.arange(0,1.01, 0.01)
        self.n_classes = n_classes
        self.reset()
        self.reduction = reduction
        self.return_roc = return_roc

    def reset(self):
        self.true_positives = torch.zeros((len(self.values), self.n_classes, ))
        self.false_positives = torch.zeros((len(self.values), self.n_classes,))
        self.all_positives = torch.zeros((self.n_classes,))
        self.all_negatives = torch.zeros((self.n_classes,))
        self.n = 0

    def update(self, batch):
        assert isinstance(batch, tuple), "batch needs to be a tuple"
        self.n += batch[0].shape[0]
        positives_mask = batch[1]==1
        negatives_mask = batch[1]==0
        self.all_positives += (positives_mask).sum(0)
        self.all_negatives += (negatives_mask).sum(0)

        for i, value in enumerate(self.values):
            prediction = torch.zeros_like(batch[0])
            prediction[batch[0] > value] = 1
            self.true_positives[i] +=  ((prediction / positives_mask) == 1).sum(0)
            self.false_positives[i] += ((prediction / positives_mask) == float("inf")).sum(0)

    def compute(self):
        if self.reduction == "micro":
            tpr = self.true_positives.sum(-1) / self.all_positives.sum()
            fpr = self.false_positives.sum(-1) / self.all_negatives.sum()

        if self.reduction == "macro":
            tpr = self.true_positives / self.all_positives[None, :]
            tpr[torch.isnan(tpr)] = 0
            fpr = self.false_positives / self.all_negatives[None, :]
            tpr = tpr.mean(-1)
            fpr = fpr.mean(-1)

        return (skm.auc(fpr, tpr), (fpr.tolist(), tpr.tolist()) )if self.return_roc else skm.auc(fpr, tpr)







import networkx as nx
import torch

def embed_align(classes, graph, model="glove50", topk=10, batch_size=50, device="cpu"):
    from ..representation import Embedder
    import re
    e = Embedder(model, device=device, return_device=device)
    classes_tokens = [" ".join(re.split("[/ _-]", x.lower())) for x in classes.keys()]

    class_embeddings = torch.stack([x.mean(-2) for x in e.embed(classes_tokens, None)],0)

    from ..representation.similarities import cos
    scores = []
    for batch in e.embed_batch_iterator(list(graph.nodes), batch_size=batch_size):
        scores.append(cos(class_embeddings, torch.stack([x.mean(-2) for x in batch],0)).t())

    scores = torch.cat(scores, 0)
    similar_nodes = scores.topk(topk, dim=0)[1].t().cpu()
    return {k:[list(graph.nodes)[x.item()] for x in v] for k,v in zip(classes.keys(), similar_nodes)}


def extend(classes, graph, depth=1, topk=20, allow_non_alignment=False, device="cpu"):
    """
    Return a subgraph of relevant nodes.
    Args:
        classes: the classes mapping
        graph:  The graph from which to draw the subgraph from
        depth: The longest path from a classes label to any node

    Returns:
    """
    augmented = embed_align(classes, graph, "glove300", topk=topk, batch_size=64, device=device)
    if not allow_non_alignment:
        assert len(augmented) == len([x for x,v in augmented.items() if len(v) > 0]), \
            "Not every class could be aligned in the graph: "+ ", ".join([x for x,v in augmented.items() if len(v) == 0])
    subgraph = nx.Graph()
    subgraph.add_nodes_from(classes.keys())
    for key, nodes in augmented.items():
        for n in nodes:
            subgraph.add_edge(key,n)

    current_nodes = list(subgraph.nodes)[len(classes):]
    for i in range(1,depth):
        next_level = {x:graph.adj[x] for x in current_nodes}
        next_nodes = []
        for u, v_dict in next_level.items():
            for v, l in v_dict.items():
                next_nodes.append(v)
                subgraph.add_edge(u,v, label=l["label"])
        current_nodes = next_nodes

    # Add all edges of nodes that are both in the subgraph already.
    for node in subgraph.nodes():
        for k, v in graph.adj.get(node,{}).items():
            if k in subgraph.nodes:
                subgraph.add_edge(node, k, label=v["label"])
    return subgraph


import networkx as nx
import torch

def subgraphs(classes, graph, depth=1, model="glove50", topk=10,  allow_non_alignment=False, batch_size=50, device="cpu"):
    from ..representation import Embedder
    import re

    e = Embedder(model, device=device, return_device=device)
    classes_tokens = [" ".join(re.split("[/ _.-]", x.lower())) for x in classes.keys()]

    class_embeddings = torch.stack([x.mean(-2) for x in e.embed(classes_tokens, None)],0)

    from ..representation.similarities import cos
    scores = []
    for batch in e.embed_batch_iterator(list(graph.nodes), batch_size=batch_size):
        scores.append(cos(class_embeddings, torch.stack([x.mean(-2) for x in batch],0)).t())

    scores = torch.cat(scores, 0)
    similar_nodes = scores.topk(topk, dim=0)[1].t().cpu()
    augmented = {k:[list(graph.nodes)[x.item()] for x in v] for k,v in zip(classes.keys(), similar_nodes)}

    if not allow_non_alignment:
        assert len(augmented) == len([x for x,v in augmented.items() if len(v) > 0]), \
            "Not every class could be aligned in the graph: "+ ", ".join([x for x,v in augmented.items() if len(v) == 0])
    subgraph = nx.DiGraph()
    subgraph.add_nodes_from(classes.keys())
    for key, nodes in augmented.items():
        for n in nodes:
            subgraph.add_edge(key,n)

    current_nodes = list(subgraph.nodes)[len(classes):]
    for i in range(1,depth):
        next_level = {x:graph.adj[x] for x in current_nodes}
        next_nodes = []
        for u, v_dict in next_level.items():
            for v, l in v_dict.items():
                next_nodes.append(v)
                subgraph.add_edge(u,v, label=l["label"])
        current_nodes = next_nodes

    # Add all edges of nodes that are both in the subgraph already.
    for node in subgraph.nodes():
        for k, v in graph.adj.get(node,{}).items():
            if k in subgraph.nodes:
                subgraph.add_edge(node, k, label=v["label"])


    matrix = e.embed(list(subgraph.nodes), pad=4)

    return subgraph, matrix


def plot_activation(graph, classes, scores, tr, target=None, title = None, layout="lgl", options={}):
    import igraph
    g = igraph.Graph(directed=True)
    g.add_vertices(list(graph.nodes()))
    g.add_edges(graph.edges())

    for x in g.vs:
        x["score"] = scores[x["name"]] -tr

    def score_color(alpha):
        if alpha <= 0:
            return igraph.color_name_to_rgb("red") + (-alpha,)
        else:
            return igraph.color_name_to_rgb("green") + (alpha,)

    import random
    random.seed(42)
    visual_style = {}
    visual_style["vertex_size"] = 5
    visual_style["vertex_color"] = [score_color(x) for x in g.vs["score"]]
    visual_style["vertex_label"] = [x["name"] if x["name"] in classes.keys() or x["score"]>0 else None for x in g.vs]
    visual_style["vertex_frame_color"] = [score_color(x)[:3] for x in g.vs["score"]]
    visual_style["vertex_frame_width"] = 0.1
    visual_style["vertex_label_size"] = 5
    visual_style["edge_width"] = 0.01
    # visual_style["edge_color"] = (0.1,0.1,0.1)
    # visual_style["label_size"] = 0.1
    # visual_style["edge_width"] = [1 + 2 * int(is_formal) for is_formal in g.es["is_formal"]]
    visual_style["layout"] =  g.layout(layout, **options)
    visual_style["bbox"] =  (1000, 1000)
    visual_style["margin"] = 20
    visual_style["edge_arrow_size"] = 0.1
    visual_style["edge_color"]=  "grey"#"(0.1,0.1,0.1)

    if title is not None:
        visual_style["main"] = title

    if target is not None:
        igraph.plot(g,target, **visual_style)
    else:
        return igraph.plot(g, **visual_style)



import torch

class Prob(torch.nn.Module):
    """Experimental feature A simple weighted vector sum.
        Idea:
            x[i, j] every activation i of the input corresponds to a distribution over the output class j
            The distributions are weighted and summed.
    """
    def __init__(self, n_labels):
        super(Prob,         self).__init__()
        self.corr = torch.nn.Parameter(torch.Tensor(n_labels, n_labels))
        torch.nn.init.eye_(self.corr)

    def forward(self, x):
        x = (torch.unsqueeze(x, -1) * self.corr).sum(-2)
        return x



import string

import torch

from .abstracts_lm import LanguageModelAbstract
from ..layers import MogrifierLSTM
from ..representation import map_vocab


class MogrifierLMCharacter(LanguageModelAbstract):
    def __init__(self, alphabet=string.ascii_letters+string.punctuation+"1234567890",
                 hidden_size=128, emb_dim=50, n_layers=1, mogrify_steps=2,
                 learn_initial_states=True, max_len=128, dropout=0.5, inter_dropout=0.2, **kwargs):
        super(MogrifierLMCharacter, self).__init__(**kwargs)


        self.max_len = max_len
        self.emb_dim = emb_dim
        self.n_layers = n_layers
        self.mogrify_steps = mogrify_steps
        self.learn_initial_states = learn_initial_states
        self.hidden_size = hidden_size
        self.cell_type = MogrifierLSTM
        self.dropout = dropout
        self.inter_dropout = inter_dropout
        self.lm_layers = torch.nn.ModuleList(
            [
                self.cell_type(hidden_size, hidden_size, mogrify_steps)
                if i > 0 else
                self.cell_type(self.emb_dim, hidden_size, mogrify_steps)
                for i in range(n_layers)]
        )
        self.dropout_layer = torch.nn.Dropout(dropout)
        self.inter_dropout_layer = torch.nn.Dropout(inter_dropout)
        self.alphabet = list(alphabet)
        self.alphabet = dict(zip(self.alphabet, range(len(self.alphabet))))
        self.vocabulary_size = len(self.alphabet)
        self.embedding = torch.nn.Embedding(
            num_embeddings=self.vocabulary_size,
            embedding_dim=self.emb_dim,
        )
        self.projection = torch.nn.Linear(self.hidden_size, self.vocabulary_size)
        self.build()

    def forward(self, x, representations=False):
        e = self.embedding(x)
        for layer in self.lm_layers:
            e = self.inter_dropout_layer(e)
            e, rep = layer(e)
        if representations:
            return e, rep[0]
        e = self.dropout_layer(e)[:, -1, :]
        return self.projection(e)

    def transform(self, s):
        return torch.tensor([[self.alphabet[x] for x in m] for m in s]).squeeze(-1)

    def encode(self, s):
        return self.transform([x for x in s if x in self.alphabet.keys()]).tolist()

    def decode(self, s):
        alphabet_rev = {v:k for k,v in self.alphabet.items()}
        return "".join([alphabet_rev[x] for x in s])

class MogrifierLMWord(LanguageModelAbstract):
    def __init__(self, word_list, hidden_size=128, emb_dim=50, n_layers=1, mogrify_steps=2,
                 learn_initial_states=True, max_len=128, dropout=0.5, inter_dropout=0.2, **kwargs):
        super(MogrifierLMWord, self).__init__(**kwargs)

        self.max_len = max_len
        self.emb_dim = emb_dim
        self.n_layers = n_layers
        self.mogrify_steps = mogrify_steps
        self.learn_initial_states = learn_initial_states
        self.hidden_size = hidden_size
        self.cell_type = MogrifierLSTM
        self.lm_layers = torch.nn.ModuleList(
            [
                self.cell_type(hidden_size, hidden_size, mogrify_steps)
                if i > 0 else
                self.cell_type(self.emb_dim, hidden_size, mogrify_steps)
                for i in range(n_layers)]
        )
        self.dropout = dropout
        self.inter_dropout = inter_dropout
        self.dropout_layer = torch.nn.Dropout(dropout)
        self.inter_dropout_layer = torch.nn.Dropout(inter_dropout)
        self.word_list = word_list + ["<UNK_TOKEN>"]
        self.word_list = dict(zip(self.word_list, range(1,len(self.word_list)+1)))
        self.vocabulary_size = len(self.word_list)+1
        self.embedding = torch.nn.Embedding(
            num_embeddings=self.vocabulary_size,
            embedding_dim=self.emb_dim,
        )
        self.projection = torch.nn.Linear(self.hidden_size, self.vocabulary_size)
        self.build()

    def forward(self, x, representations=False):
        e = self.embedding(x)
        for layer in self.lm_layers:
            e = self.inter_dropout_layer(e)
            e, rep = layer(e)
        if representations:
            return e, rep[0]
        e = self.dropout_layer(e)[:, -1, :]
        return self.projection(e)

    def transform(self, s):
        if isinstance(s[0], str):
            s = [s]
        return map_vocab(s,self.word_list,len(s[0])).t().squeeze(-1)

    def encode(self, s):
        return self.transform([s.split()]).tolist()

    def decode(self, s):
        return " ".join([list(self.word_list.keys())[x-1] for x in s])



from ignite.metrics import Precision, Accuracy
import torch


class PrecisionK(Precision):
    def __init__(self, k, *args, **kwargs):
        self.k = k
        super(PrecisionK, self).__init__(*args, **kwargs)

    def update(self, output):
        transformed = torch.zeros_like(output[0]).scatter(1, torch.topk(output[0], k=self.k)[1], 1)
        super(PrecisionK, self).update((transformed, output[1]))

class AccuracyTreshold(Accuracy):
    def __init__(self, trf, args_dict={}, *args, **kwargs):
        self.trf = trf
        self.args_dict = args_dict
        super(AccuracyTreshold, self).__init__(*args, **kwargs)

    def update(self, output):
        super(AccuracyTreshold, self).update((self.trf(x=output[0], **self.args_dict), output[1]))
